# -*- coding: utf-8 -*-
import numpy as np
from dipy.reconst.shm import real_sph_harm
from dipy.utils.optpkg import optional_package
numba, have_numba, _ = optional_package("numba")

__all__ = [
    'real_sym_rh_basis',
    'sh_convolution'
]


def real_sym_rh_basis(sh_order, theta, phi):
    r"""Samples a real symmetric rotational harmonic basis at point on the sphere

    Samples the basis functions up to order `sh_order` at points on the sphere
    given by `theta` and `phi`. The basis functions are defined here the same
    way as in fibernavigator, where the real harmonic $Y^m_n$ is defined to
    be:

        $Y^0_n$                     if m = 0

    Parameters
    -----------
    sh_order : int
        even int > 0, max spherical harmonic degree
    theta : float [0, 2*pi]
        The azimuthal (longitudinal) coordinate.
    phi : float [0, pi]
        The polar (colatitudinal) coordinate.

    Returns
    --------
    real_rh_matrix : array of shape ()
        The real harmonic $Y^0_n$ sampled at `theta` and `phi`
    """
    n = np.arange(0, sh_order + 1, 2)
    m = np.zeros(sh_order // 2 + 1)

    phi = np.reshape(phi, [-1, 1])
    theta = np.reshape(theta, [-1, 1])

    real_rh_matrix = real_sph_harm(m, n, theta, phi)
    return real_rh_matrix


def sh_convolution(f_distribution_sh, kernel_rh):
    r"""Spherical convolution between a fiber distribution (f) in spherical
    harmonics and a kernel in terms of rotational harmonics (oriented along the
    z-axis).

    Parameters
    ----------
    f_distribution_sh : array, shape (sh_coef)
        spherical harmonic coefficients of a fiber distribution.
    kernel_rh : array, shape (sh_coef),
        rotational harmonic coefficients of the convolution kernel. In our case
        this is the spherical signal of one micro-environment at one b-value.

    Returns
    -------
    f_kernel_convolved : array, shape (sh_coef)
        spherical harmonic coefficients of the convolved kernel and
        distribution.
    """
    sh_order_rh = 2 * (len(kernel_rh) - 1)
    number_coef_sh = len(f_distribution_sh)
    sh_order_sh = int(-3 + np.sqrt(9 - 4 * (2 - 2 * number_coef_sh))) // 2

    sh_order_used = min(sh_order_rh, sh_order_sh)
    number_coef_used = int((sh_order_used + 2) * (sh_order_used + 1) // 2)

    f_kernel_convolved = np.zeros(number_coef_used)

    counter = 0
    for n_ in range(0, sh_order_used + 1, 2):
        coef_in_order = 2 * n_ + 1
        f_kernel_convolved[counter: counter + coef_in_order] = (
            f_distribution_sh[counter: counter + coef_in_order] *
            kernel_rh[n_ // 2] * np.sqrt((4 * np.pi) / (2 * n_ + 1))
        )
        counter += coef_in_order
    return f_kernel_convolved


if have_numba:
    sh_convolution = numba.njit()(sh_convolution)



import numpy as np
from dipy.data import get_sphere, HemiSphere
from dipy.reconst.shm import real_sym_sh_mrtrix
from dipy.utils.optpkg import optional_package
from dipy.reconst.shm import sph_harm_ind_list
cvxpy, have_cvxpy, _ = optional_package("cvxpy")
sphere = get_sphere('symmetric724')


__all__ = [
    'CsdCvxpyOptimizer'
]


class CsdCvxpyOptimizer:
    """
    Generalized multi-compartment constrained spherical deconvolution (MC-CSD)
    optimizer. The algorithm follows the formulation of Multi-Tissue (MT)-CSD
    [1]_, but is completely generalized in that it can take any number or type
    of convolution kernels, have static or voxel-varying kernels parameters,
    and can have fixed volume fractions or estimate them. The algorithm is
    implemented using CVXPY [2]_.

    Limitations:
    - It cannot estimate the volume fractions of multiple kernels that each
      have an orientation. E.g. it is possible to have a cylinder and a ball
      kernel as input, but not two cylinders.
    - Currently, the different compartment kernels are directly fitted to the
      signal attenuation, implicitly assuming each compartment has the same
      b0-intensity. True MT-CSD requires that these B0-intensities are also
      included.

    Parameters
    ----------
    acquisition_scheme: DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using Dmipy.
    model: Dmipy MultiCompartmentSphericalHarmonicsModel instance,
        Contains Dmipy multi-compartment model information.
    x0_vector: Array of size (Nparameters)
        Possible parameters for model kernels.
    sh_order: positive even integer,
        Spherical harmonics order for deconvolution.
    unity_constraint: boolean,
        whether or not to constrain the volume fractions of the FOD to
        unity. In the case of one model this means that the SH-coefficients
        represent a distribution on the sphere.
    lambda_lb: positive float,
        Laplace-Belrami regularization weight to impose smoothness in the
        FOD. Same as is done in [3]_.

    References
    ----------
    .. [1] Jeurissen, Ben, et al. "Multi-tissue constrained spherical
        deconvolution for improved analysis of multi-shell diffusion MRI
        data." NeuroImage 103 (2014): 411-426.
    .. [2] Diamond, Steven, and Stephen Boyd. "CVXPY: A Python-embedded
        modeling language for convex optimization." The Journal of Machine
        Learning Research 17.1 (2016): 2909-2913.
    .. [3] Descoteaux, Maxime, et al. "Regularized, fast, and robust analytical
        Q-ball imaging." Magnetic Resonance in Medicine: An Official Journal of
        the International Society for Magnetic Resonance in Medicine 58.3
        (2007): 497-510.
    """

    def __init__(self, acquisition_scheme, model, x0_vector=None, sh_order=8,
                 unity_constraint=True, lambda_lb=0.):
        self.model = model
        self.acquisition_scheme = acquisition_scheme
        self.sh_order = sh_order
        self.Ncoef = int((sh_order + 2) * (sh_order + 1) // 2)
        self.Nmodels = len(self.model.models)
        self.lambda_lb = lambda_lb
        self.unity_constraint = unity_constraint
        self.sphere_jacobian = 2 * np.sqrt(np.pi)

        sphere = get_sphere('symmetric724')
        hemisphere = HemiSphere(phi=sphere.phi, theta=sphere.theta)
        self.L_positivity = real_sym_sh_mrtrix(
            self.sh_order, hemisphere.theta, hemisphere.phi)[0]

        x0_single_voxel = np.reshape(
            x0_vector, (-1, x0_vector.shape[-1]))[0]
        if np.all(np.isnan(x0_single_voxel)):
            self.single_convolution_kernel = True
            parameters_dict = self.model.parameter_vector_to_parameters(
                x0_single_voxel)
            self.A = self.model._construct_convolution_kernel(
                **parameters_dict)
        else:
            self.single_convolution_kernel = False

        self.Ncoef_total = 0
        vf_array = []

        if self.model.volume_fractions_fixed:
            self.sh_start = 0
            self.Ncoef_total = self.Ncoef
            self.vf_indices = np.array([0])
        else:
            for model in self.model.models:
                if 'orientation' in model.parameter_types.values():
                    self.sh_start = self.Ncoef_total
                    sh_model = np.zeros(self.Ncoef)
                    sh_model[0] = 1
                    vf_array.append(sh_model)
                    self.Ncoef_total += self.Ncoef
                else:
                    vf_array.append(1)
                    self.Ncoef_total += 1
            self.vf_indices = np.where(np.hstack(vf_array))[0]

        sh_l = sph_harm_ind_list(sh_order)[1]
        lb_weights = sh_l ** 2 * (sh_l + 1) ** 2  # laplace-beltrami [3]
        if self.model.volume_fractions_fixed:
            self.R_smoothness = np.diag(lb_weights)
        else:
            diagonal = np.zeros(self.Ncoef_total)
            diagonal[self.sh_start: self.sh_start + self.Ncoef] = lb_weights
            self.R_smoothness = np.diag(diagonal)

    def __call__(self, data, x0_vector):
        """
        The fitting function of Multi-Compartment CSD optimizer.

        If there is only one convolution kernel, it loads the precalculated
        values. If the kernel is voxel-varying, it calculates it now.

        Adds a constraint that the volume fractions add up to one depending on
        the unity_constraint being True or False.

        Adds Laplace-Beltrami (smoothness) regularization if lambda_lb>0.

        Parameters
        ----------
        data : Array of size (Ndata)
            The normalized dMRI signal attenuation to be fitted.
        x0_vector : Array of size (Nparameters)
            Possible initial guess vector for parameter initiation.

        Returns
        -------
        fitted_parameter_vector : Array of size (Nparameters),
            The fitted MC model parameters using CSD.
        """

        if self.single_convolution_kernel:
            A = self.A
        else:
            parameters_dict = self.model.parameter_vector_to_parameters(
                x0_vector)
            A = self.model._construct_convolution_kernel(**parameters_dict)

        sh_coef = cvxpy.Variable(self.Ncoef_total)
        sh_fod = sh_coef[self.sh_start: self.Ncoef + self.sh_start]

        constraints = []
        constraints.append(
            self.L_positivity * sh_fod >= 0)
        vf = sh_coef[self.vf_indices] * self.sphere_jacobian
        constraints.append(vf >= 0)
        if self.unity_constraint:
            constraints.append(cvxpy.sum(vf) == 1.)

        # fix volume fractions if only some of them are fixed.
        # not if all of them are fixed - in that case the convolution
        # matrix is joined into a single composite response function.
        if not self.model.volume_fractions_fixed:
            params = self.model.parameter_vector_to_parameters(x0_vector)
            params = self.model.add_linked_parameters_to_parameters(params)
            for i, vf_name in enumerate(self.model.partial_volume_names):
                if not self.model.parameter_optimization_flags[vf_name]:
                    constraints.append(vf[i] == params[vf_name])

        cost = cvxpy.sum_squares(A * sh_coef - data)
        if self.lambda_lb > 0:
            cost += (
                self.lambda_lb * cvxpy.quad_form(sh_coef, self.R_smoothness))
        problem = cvxpy.Problem(cvxpy.Minimize(cost), constraints)
        try:
            problem.solve()
        except cvxpy.error.SolverError:
            msg = 'cvxpy solver failed'
            print(msg)
            return np.zeros_like(x0_vector)

        if problem.status in ["infeasible", "unbounded"]:
            msg = 'cvxpy found {} problem'.format(problem.status)
            print(msg)
            return np.zeros_like(x0_vector)

        # return optimized fod sh coefficients
        fitted_params = self.model.parameter_vector_to_parameters(x0_vector)
        fitted_params['sh_coeff'] = np.array(sh_fod.value).squeeze()
        # explicitly set zeroth coefficient to inverse of jacobian
        # to separate volume and shape of coefficients
        fitted_params['sh_coeff'][0] = 1. / self.sphere_jacobian

        if not self.model.volume_fractions_fixed:  # if vf was estimated
            fractions_array = np.array(
                sh_coef[self.vf_indices].value).squeeze() * 2 * np.sqrt(np.pi)
            for i, name in enumerate(self.model.partial_volume_names):
                fitted_params[name] = fractions_array[i]
        fitted_parameter_vector = self.model.parameters_to_parameter_vector(
            **fitted_params)
        return fitted_parameter_vector



import numpy as np
from os.path import join
import os
import pkg_resources
from ..core.acquisition_scheme import (
    acquisition_scheme_from_bvalues,
    acquisition_scheme_from_gradient_strengths,
    acquisition_scheme_from_schemefile)
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
import zipfile

_GRADIENT_TABLES_PATH = pkg_resources.resource_filename(
    'dmipy', 'data/gradient_tables'
)
DATA_PATH = pkg_resources.resource_filename(
    'dmipy', 'data'
)


__all__ = [
    'wu_minn_hcp_acquisition_scheme',
    'duval_cat_spinal_cord_2d_acquisition_scheme'
]


def wu_minn_hcp_acquisition_scheme():
    "Returns DmipyAcquisitionScheme of Wu-Minn HCP project."
    _bvals = np.loadtxt(
        join(_GRADIENT_TABLES_PATH,
             'bvals_hcp_wu_minn.txt')
    ) * 1e6
    _gradient_directions = np.loadtxt(
        join(_GRADIENT_TABLES_PATH,
             'bvecs_hcp_wu_minn.txt')
    )
    _delta = 0.0106
    _Delta = 0.0431
    return acquisition_scheme_from_bvalues(
        _bvals, _gradient_directions, _delta, _Delta)


def duval_cat_spinal_cord_2d_acquisition_scheme():
    "Returns 2D DmipyAcquisitionScheme of cat spinal cord data."
    scheme_name = 'tanguy_cat_spinal_cord/2D_qspace.scheme'
    scheme = np.loadtxt(join(DATA_PATH, scheme_name), skiprows=3)

    bvecs = scheme[:, :3]
    bvecs[np.linalg.norm(bvecs, axis=1) == 0.] = np.r_[1., 0., 0.]
    G = scheme[:, 3]
    Delta = scheme[:, 4]
    delta = scheme[:, 5]
    TE = scheme[:, 6]

    return acquisition_scheme_from_gradient_strengths(
        G, bvecs, delta, Delta, TE)


def duval_cat_spinal_cord_3d_acquisition_scheme():
    "Returns 3D DmipyAcquisitionScheme of cat spinal cord data."
    scheme_name = 'tanguy_cat_spinal_cord/3D_qspace.scheme'
    scheme = np.loadtxt(join(DATA_PATH, scheme_name), skiprows=3)

    bvecs = scheme[:, :3]
    bvecs[np.linalg.norm(bvecs, axis=1) == 0.] = np.r_[1., 0., 0.]
    G = scheme[:, 3]
    Delta = scheme[:, 4]
    delta = scheme[:, 5]
    TE = scheme[:, 6]
    # there's very small differences in TE between shells that we ignore
    TE[:] = 0.0472
    return acquisition_scheme_from_gradient_strengths(
        G, bvecs, delta, Delta, TE, min_b_shell_distance=20e6)


def de_santis_generated_acquisition_scheme():
    "Returns 2D DmipyAcquisitionScheme of de Santis."
    scheme_name = 'de_santis_camino/deSantis_scheme.scheme'
    scheme = np.loadtxt(join(DATA_PATH, scheme_name), skiprows=2)

    bvecs = scheme[:, :3]
    bvecs[np.linalg.norm(bvecs, axis=1) == 0.] = np.r_[1., 0., 0.]
    G = scheme[:, 3]
    Delta = scheme[:, 4]
    delta = scheme[:, 5]
    TE = scheme[:, 6]

    return acquisition_scheme_from_gradient_strengths(
        G, bvecs, delta, Delta, TE)


def isbi2015_white_matter_challenge_scheme():
    "Returns 35-shell ISBI 2015 challenge DmipyAcquisitionScheme."
    isbi_data_path = join(DATA_PATH, 'isbi2015_white_matter_challenge')
    filename = 'isbi_schemefile.txt'
    return acquisition_scheme_from_schemefile(join(isbi_data_path, filename))


def panagiotaki_verdict_acquisition_scheme():
    "Returns acquisition scheme for VERDICT tumor characterization."
    verdict_data_path = join(DATA_PATH, 'panagiotaki_verdict')
    if not os.path.exists(verdict_data_path):
        os.makedirs(verdict_data_path)

    url = "http://camino.cs.ucl.ac.uk/uploads/Tutorials/"
    schemename = "VC_DTIDW.scheme.zip"
    response = urlopen(join(url, schemename))
    with open(join(verdict_data_path, schemename), 'wb') as f:
        f.write(response.read())
    with zipfile.ZipFile(join(verdict_data_path, schemename)) as zip:
        zip.extract("VC_DTIDW.scheme", path=verdict_data_path)

    return acquisition_scheme_from_schemefile(
        join(verdict_data_path, "VC_DTIDW.scheme"), b0_threshold=5e6)



import time
import numpy as np
from dmipy.core.modeling_framework import (
    MultiCompartmentSphericalHarmonicsModel)
from dmipy.tissue_response.three_tissue_response import (
    three_tissue_response_dhollander16)


def single_shell_three_tissue_csd(
        acquisition_scheme, data, tissue_response_models=None,
        S0_tissue_responses=None, N_iterations=4, return_all_csd_fits=False,
        csd_fit_verbose=False, ss3t_verbose=True):
    """
    Implementation of Single-Shell (1 shell + b0) three-tissue CSD algorithm
    [1]_. The algorithm is based on a biconvex optimization strategy that is
    formulated as two steps that repeat:

    for k in iterations:
        # step 1
        if k == 0:
            fix WM fraction to 0 and optimize CSF + GM.
            this leads to an overestimation of GM and underestimation of CSF.
        else:
            fix WM to WM estimate of second step of last iteration.
        # step 2
        Fix CSF and fit WM + GM.
        Since CSF is an underestimate, WM will be as well.

    The authors of [1]_ don't really elaborate on this, but this algorithm will
    just slowly converge to the naive optimization approach of just fitting
    WM + GM + CSF together. The trick is that by forcing a particular starting
    point for the biconvex optimization (WM=0), typically after around 4
    iterations of this algorithm, the fractions will *temporarily* have values
    that are quite close to the multi-shell multi-tissue algorithm.

    This algorithm seems to work better for higher b-value HARDI data than for
    DTI data when it comes to GM fractions.

    Parameters
    ----------
    acquisition_scheme: Dmipy acquisition scheme,
        single shell (or whatever-shell) acquisition scheme.
    data: ND-array of shape (Nx...., NDWI),
        the fitted that is to be fitted.
    tissue_response_models: list of Dmipy tissue response models,
        assumed to be in the same order as they are generated using the
        dhollander16 tissue response estimation, i.e. [wm, gm, csf].
        if not given it is estimated from the data using the Dhollander16
        heuristic three-tissue estimation algorithm [2]_.
    S0_tissue_responses: list of floats,
        S0 response values corresponding to input tissue response models.
    N_iterations: positive integer,
        number of biconvex optimization steps to do in the algorithm.
        default is 4 according to [1]_.
    return_all_csd_fits: bool,
        whether to return all csd fits of all iterations or just the last one.
    csd_fit_verbose: bool,
        whether to suppress csd fitting prints or not.
    ss3t_verbose: bool,
        whether to suppress the single-shell three tissue prints or not.

    Returns
    -------
    mt_csd_fits or mt_csd_fit: list of, or singular,
        FittedMultiCompartmentSphericalHarmonicsModel.

    References
    ----------
    .. [1] Dhollander, Thijs, and Alan Connelly. "A novel iterative approach to
        reap the benefits of multi-tissue CSD from just single-shell (+ b= 0)
        diffusion MRI data." 24th International Society of Magnetic Resonance
        in Medicine 24 (2016): 3010.
    .. [2] Dhollander, T.; Raffelt, D. & Connelly, A. Unsupervised 3-tissue
        response function estimation from single-shell or multi-shell diffusion
        MR data without a co-registered T1 image. ISMRM Workshop on Breaking
        the Barriers of Diffusion MRI, 2016, 5
    """
    if tissue_response_models is None:
        S0_tissue_responses, tissue_response_models, selection_map = (
            three_tissue_response_dhollander16(acquisition_scheme, data))

    fit_args = {
        'acquisition_scheme': acquisition_scheme,
        'data': data,
        'mask': data[..., 0] > 0,
        'fit_S0_response': True,
        'verbose': csd_fit_verbose}

    if return_all_csd_fits:
        mt_csd_fits = []
    for it in range(N_iterations):
        start = time.time()
        # step one: fix WM and fit GM + CSF
        mt_csd_mod = MultiCompartmentSphericalHarmonicsModel(
            tissue_response_models, S0_tissue_responses)
        if it == 0:
            mt_csd_mod.set_fixed_parameter(
                'partial_volume_0',
                np.zeros(data.shape[:-1]))
        else:
            mt_csd_mod.set_fixed_parameter(
                'partial_volume_0',
                mt_csd_fit.fitted_parameters['partial_volume_0'])  # noqa: F821
        mt_csd_fit = mt_csd_mod.fit(**fit_args)
        if return_all_csd_fits:
            mt_csd_fits.append(mt_csd_fit)

        # step two: fix CSF and fit WM + GM
        mt_csd_mod = MultiCompartmentSphericalHarmonicsModel(
            tissue_response_models, S0_tissue_responses)
        mt_csd_mod.set_fixed_parameter(
            'partial_volume_2',
            mt_csd_fit.fitted_parameters['partial_volume_2'])
        mt_csd_fit = mt_csd_mod.fit(**fit_args)
        if return_all_csd_fits:
            mt_csd_fits.append(mt_csd_fit)
        computation_time = time.time() - start
        if ss3t_verbose:
            print('finish it {} of {} in {} seconds'.format(
                it + 1, N_iterations, int(computation_time)))
    if return_all_csd_fits:
        return mt_csd_fits
    else:
        return mt_csd_fit



from os.path import join
import os
import pkg_resources
import nibabel as nib
import numpy as np
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
import zipfile
from . import saved_acquisition_schemes

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

DATA_PATH = pkg_resources.resource_filename(
    'dmipy', 'data'
)

__all__ = [
    'wu_minn_hcp_coronal_slice',
    'duval_cat_spinal_cord_2d',
    'synthetic_camino_data_parallel',
    'synthetic_camino_data_dispersed',
    'visualize_correlation_camino_and_estimated_fractions'
]


def de_santis_camino_data():
    """
    Downloads and returns the 4-shell multi-delta/Delta/G scheme based on
    acquistion scheme defined in [1]_. Note that acquisition parameters in [1]_
    used for a STEAM sequence, are used here to generate a PGSE one.

    Returns
    -------
    scheme: DmipyAcquisitionScheme instance,
        acquisition scheme of the generated de Santis data.
    data_genu: array of size (50, 54),
        contains 50 repetitions with added rician noise SNR=30.

    References
    ----------
    .. [1] De Santis, S., Jones, D. K., & Roebroeck, A. (2016). Including
        diffusion time dependence in the extra-axonal space improves in vivo
        estimates of axonal diameter and density in human white matter.
        NeuroImage, 130, 91-103.
    """
    de_santis_data_path = join(DATA_PATH, 'de_santis_camino')

    data = np.loadtxt(join(de_santis_data_path, 'deSantis_signal.txt'))

    scheme = (
        saved_acquisition_schemes.de_santis_generated_acquisition_scheme())
    return scheme, data


def wu_minn_hcp_coronal_slice():
    "Returns example slice of Wu-Minn HCP data subject 100307."
    data_path = join(
        DATA_PATH, 'hcp', 'hcp_example_slice', 'coronal_slice.nii.gz')
    try:
        data = nib.load(data_path).get_data()
    except IOError:
        msg = "The example HCP data has not been downloaded yet. "
        msg += "Please follow our HCP tutorial where you can use your own AWS "
        msg += "credentials to download the example data and any other HCP "
        msg += "subject data."
        raise ValueError(msg)

    scheme = saved_acquisition_schemes.wu_minn_hcp_acquisition_scheme()

    msg = "This data slice originates from Subject 100307 of the Human "
    msg += "Connectome Project, WU-Minn Consortium (Principal Investigators: "
    msg += "David Van Essen and Kamil Ugurbil; 1U54MH091657) funded by the 16 "
    msg += "NIH Institutes and Centers that support the NIH Blueprint for "
    msg += "Neuroscience Research; and by the McDonnell Center for Systems "
    msg += "Neuroscience at Washington University."
    print(msg)

    return scheme, data


def duval_cat_spinal_cord_2d():
    "Returns 2D multi-diffusion time AxCaliber data of cat spinal cord."
    msg = "This data was used by Duval et al. 'Validation of quantitative MRI "
    msg += "metrics using full slice histology with automatic axon "
    msg += "segmentation', ISMRM 2016. Reference at "
    msg += "Cohen-Adad et al. White Matter Microscopy Database."
    msg += " http://doi.org/10.17605/OSF.IO/YP4QG"
    print(msg)

    data_folder = join(DATA_PATH, "tanguy_cat_spinal_cord")

    class Histology:
        def __init__(self):
            self.h1_axonEquivDiameter = nib.load(
                join(data_folder, '1_axonEquivDiameter.nii')).get_data()
            self.h2_axonEquivDiameter_std = nib.load(
                join(data_folder, '2_axonEquivDiameter_std.nii')).get_data()
            self.h3_axonEquivDiameter_axonvolumeCorrected = nib.load(
                join(data_folder, '3_axonEquivDiameter_axonvolumeCorrected.nii'
                     )).get_data()
            self.h4_fr = nib.load(join(data_folder, '4_fr.nii')).get_data()
            self.h5_MyelinVolumeFraction = nib.load(
                join(data_folder, '5_MyelinVolumeFraction.nii')).get_data()
            self.h6_gRatio = nib.load(
                join(data_folder, '6_gRatio.nii')).get_data()
            self.h7_Number_axons = nib.load(
                join(data_folder, '7_Number_axons.nii')).get_data()

    class DuvalSpinalCordData2D:
        def __init__(self):
            data_name = "tanguy_spinal_cord_2D.nii.gz"
            self.signal = nib.load(join(data_folder, data_name)).get_data()
            self.histology = Histology()
            self.mask = (self.histology.h4_fr > 0)[..., None]

    data = DuvalSpinalCordData2D()
    scheme = (
        saved_acquisition_schemes.duval_cat_spinal_cord_2d_acquisition_scheme()
    )
    return scheme, data


def duval_cat_spinal_cord_3d():
    "Returns 2D multi-diffusion time AxCaliber data of cat spinal cord."
    msg = "This data was used by Duval et al. 'Validation of quantitative MRI "
    msg += "metrics using full slice histology with automatic axon "
    msg += "segmentation', ISMRM 2016. Reference at "
    msg += "Cohen-Adad et al. White Matter Microscopy Database."
    msg += " http://doi.org/10.17605/OSF.IO/YP4QG"
    print(msg)

    data_folder = join(DATA_PATH, "tanguy_cat_spinal_cord")

    class Histology:
        def __init__(self):
            self.h1_axonEquivDiameter = nib.load(
                join(data_folder, '1_axonEquivDiameter.nii')).get_data()
            self.h2_axonEquivDiameter_std = nib.load(
                join(data_folder, '2_axonEquivDiameter_std.nii')).get_data()
            self.h3_axonEquivDiameter_axonvolumeCorrected = nib.load(
                join(data_folder, '3_axonEquivDiameter_axonvolumeCorrected.nii'
                     )).get_data()
            self.h4_fr = nib.load(join(data_folder, '4_fr.nii')).get_data()
            self.h5_MyelinVolumeFraction = nib.load(
                join(data_folder, '5_MyelinVolumeFraction.nii')).get_data()
            self.h6_gRatio = nib.load(
                join(data_folder, '6_gRatio.nii')).get_data()
            self.h7_Number_axons = nib.load(
                join(data_folder, '7_Number_axons.nii')).get_data()

    class DuvalSpinalCordData3D:
        def __init__(self):
            data_name = "tanguy_spinal_cord_3D.nii.gz"
            self.signal = nib.load(join(data_folder, data_name)).get_data()
            self.histology = Histology()
            self.mask = (self.histology.h4_fr > 0)[..., None]

    data = DuvalSpinalCordData3D()
    scheme = (
        saved_acquisition_schemes.duval_cat_spinal_cord_3d_acquisition_scheme()
    )
    return scheme, data


def isbi2015_white_matter_challenge():
    """
    Downloads and returns the 35-shell multi-delta/Delta/G scheme and data for
    the fornix and genu data that was used for the ISBI 2015 white matter
    challenge [1]_.

    Returns
    -------
    scheme: DmipyAcquisitionScheme instance,
        acquisition scheme of the challenge data.
    data_genu: array of size (3612, 6),
        contains the DWIs for 6 genu voxels.
    data_fornix: array of size (3612, 6),
        contains the DWIs for 6 fornix voxels.

    References
    ----------
    .. [1] Ferizi, Uran, et al. "Diffusion MRI microstructure models with in
        vivo human brain Connectome data: results from a multi-group
        comparison." NMR in Biomedicine 30.9 (2017)
    """
    isbi_data_path = join(DATA_PATH, 'isbi2015_white_matter_challenge')
    data_genu = np.loadtxt(join(isbi_data_path, 'genu.txt'), skiprows=1)
    data_fornix = np.loadtxt(join(isbi_data_path, 'fornix.txt'), skiprows=1)
    scheme = (
        saved_acquisition_schemes.isbi2015_white_matter_challenge_scheme()
    )
    return scheme, data_genu.T, data_fornix.T


def panagiotaki_verdict():
    """
    Downloads and returns the example VERDICT acquisition scheme and data that
    is available at the UCL website. The data is an example of [1]_.

    Returns
    -------
    scheme: DmipyAcquisitionScheme instance,
        acquisition scheme of the challenge data.
    data_verdict: array,
        contains the DWIs for a single tumor voxel.

    References
    ----------
    .. [1] Panagiotaki, Eletheria, et al. "Noninvasive quantification of solid
        tumor microstructure using VERDICT MRI." Cancer research 74.7 (2014):
        1902-1912.
    """
    verdict_data_path = join(DATA_PATH, 'panagiotaki_verdict')
    if not os.path.exists(verdict_data_path):
        os.makedirs(verdict_data_path)

    url = "http://camino.cs.ucl.ac.uk/uploads/Tutorials/"
    filename = "LSDTIDWtut.Bfloat.zip"
    response = urlopen(join(url, filename))
    with open(join(verdict_data_path, filename), 'wb') as f:
        f.write(response.read())
    with zipfile.ZipFile(join(verdict_data_path, filename)) as zip:
        zip.extract("LSDTIDWtut.Bfloat", path=verdict_data_path)

    data_verdict = np.fromfile(join(verdict_data_path, "LSDTIDWtut.Bfloat"),
                               dtype='>f')
    scheme = saved_acquisition_schemes.panagiotaki_verdict_acquisition_scheme()
    return scheme, data_verdict


def synthetic_camino_data_parallel():
    """The parallel data was generated using the Camino Monte-Carlo
    Diffusion Simulator. See http://camino.cs.ucl.ac.uk/.
    """
    fractions_1_7 = np.loadtxt(
        join(DATA_PATH, 'camino', 'fractions_camino_D1_7.txt'))
    fractions_2_0 = np.loadtxt(
        join(DATA_PATH, 'camino', 'fractions_camino_D2_0.txt'))
    fractions_2_3 = np.loadtxt(
        join(DATA_PATH, 'camino', 'fractions_camino_D2_3.txt'))

    data_1_7 = np.loadtxt(join(DATA_PATH, 'camino', 'data_camino_D1_7.txt'))
    data_2_0 = np.loadtxt(join(DATA_PATH, 'camino', 'data_camino_D2_0.txt'))
    data_2_3 = np.loadtxt(join(DATA_PATH, 'camino', 'data_camino_D2_3.txt'))

    fractions = np.r_[fractions_1_7, fractions_2_0, fractions_2_3]
    data = np.r_[data_1_7, data_2_0, data_2_3]
    diffusivity = np.r_[np.tile(1.7e-9, len(fractions_1_7)),
                        np.tile(2e-9, len(fractions_2_0)),
                        np.tile(2.3e-9, len(fractions_2_3))]

    class CaminoData:
        def __init__(self):
            self.fractions = fractions
            self.diffusivities = diffusivity
            self.signal_attenuation = data

    scheme = saved_acquisition_schemes.wu_minn_hcp_acquisition_scheme()
    return scheme, CaminoData()


def synthetic_camino_data_dispersed():
    """The dispersed data was generated by using the parallel Camino data as
    an described above, and then dispersing it using Watson and Bingham
    distributions.
    """
    data_1_7_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'data_camino_dispersed_D1_7.txt'))
    data_2_0_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'data_camino_dispersed_D2_0.txt'))
    data_2_3_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'data_camino_dispersed_D2_3.txt'))
    data = np.r_[
        data_1_7_dispersed, data_2_0_dispersed, data_2_3_dispersed]

    parameters_1_7_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'parameters_camino_dispersed_D1_7.txt'))
    parameters_2_0_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'parameters_camino_dispersed_D2_0.txt'))
    parameters_2_3_dispersed = np.loadtxt(
        join(DATA_PATH, 'camino', 'parameters_camino_dispersed_D2_3.txt'))

    fractions = np.r_[
        parameters_1_7_dispersed[:, 0],
        parameters_2_0_dispersed[:, 0],
        parameters_2_3_dispersed[:, 0]
    ]

    kappas = np.r_[
        parameters_1_7_dispersed[:, 1],
        parameters_2_0_dispersed[:, 1],
        parameters_2_3_dispersed[:, 1]
    ]

    betas = np.r_[
        parameters_1_7_dispersed[:, 2],
        parameters_2_0_dispersed[:, 2],
        parameters_2_3_dispersed[:, 2]
    ]

    diffusivity = np.r_[np.tile(1.7e-9, len(parameters_1_7_dispersed)),
                        np.tile(2e-9, len(parameters_2_0_dispersed)),
                        np.tile(2.3e-9, len(parameters_2_3_dispersed))]

    class DispersedCaminoData:
        def __init__(self):
            self.fractions = fractions
            self.diffusivities = diffusivity
            self.signal_attenuation = data
            self.kappa = kappas
            self.beta = betas

    scheme = saved_acquisition_schemes.wu_minn_hcp_acquisition_scheme()
    return scheme, DispersedCaminoData()


def visualize_correlation_camino_and_estimated_fractions(
        estim_fractions_parallel, estim_fractions_dispersed):
    "Function that visualizes Camino estimated results versus ground truth."

    data_parallel = synthetic_camino_data_parallel()
    data_dispersed = synthetic_camino_data_dispersed()

    mask_par_17 = data_parallel.diffusivities == 1.7e-9
    mask_disp_17 = data_dispersed.diffusivities == 1.7e-9

    fractions_par_17 = data_parallel.fractions[mask_par_17]
    fractions_disp_17 = data_dispersed.fractions[mask_disp_17]

    estim_fractions_par_17 = estim_fractions_parallel[mask_par_17]
    estim_fractions_disp_17 = estim_fractions_dispersed[mask_disp_17]

    pr = pearsonr(estim_fractions_par_17, fractions_par_17)
    pr_dispersed = pearsonr(estim_fractions_disp_17, fractions_disp_17)
    pr_multidif = pearsonr(estim_fractions_parallel, data_parallel.fractions)
    pr_multidif_dispersed = pearsonr(
        estim_fractions_dispersed, data_dispersed.fractions)

    f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
        2, 2, sharex='col', sharey='row')
    ax1.scatter(fractions_par_17, estim_fractions_par_17)
    ax2.scatter(data_parallel.fractions, estim_fractions_parallel)
    ax3.scatter(fractions_disp_17, estim_fractions_disp_17)
    ax4.scatter(data_dispersed.fractions, estim_fractions_dispersed)

    ax1.text(.216, .817, 'pearsonR= ' +
             str(np.round(pr[0], 3)), fontsize=10,
             bbox=dict(facecolor='white', alpha=1))
    ax2.text(.216, .817, 'pearsonR= ' +
             str(np.round(pr_multidif[0], 3)), fontsize=10,
             bbox=dict(facecolor='white', alpha=1))
    ax3.text(.216, .817, 'pearsonR= ' +
             str(np.round(pr_dispersed[0], 3)), fontsize=10,
             bbox=dict(facecolor='white', alpha=1))
    ax4.text(.216, .817, 'pearsonR= ' + str(np.round(
        pr_multidif_dispersed[0], 3)), fontsize=10,
        bbox=dict(facecolor='white', alpha=1))

    ax1.set_title('Static Diffusivity')
    ax3.set_xlabel('Ground Truth')
    ax2.set_title('Varying Diffusivity')
    ax1.set_ylabel('Estimated intra-vf')
    ax4.set_xlabel('Ground Truth')
    ax3.set_ylabel('Estimated intra-vf')

    ax1.plot([0, 1], [0, 1], ls='--', c='k', lw=3)
    ax2.plot([0, 1], [0, 1], ls='--', c='k', lw=3)
    ax3.plot([0, 1], [0, 1], ls='--', c='k', lw=3)
    ax4.plot([0, 1], [0, 1], ls='--', c='k', lw=3)
    ax1.set_ylim(0.2, .9)
    ax1.set_xlim(0.2, .8)
    ax4.set_ylim(0.2, .9)
    ax4.set_xlim(0.2, .8)



import warnings
from scipy.optimize import brute
from scipy.stats import pearsonr
import numpy as np
from dipy.reconst import dti
from ..core.acquisition_scheme import gtab_dmipy2dipy
from dipy.segment.mask import median_otsu
from . import white_matter_response
from ..signal_models.tissue_response_models import (
    estimate_TR1_isotropic_tissue_response_model)

_white_matter_response_algorithms = {
    'tournier07': white_matter_response.white_matter_response_tournier07,
    'tournier13': white_matter_response.white_matter_response_tournier13
}


def three_tissue_response_dhollander16(
        acquisition_scheme, data, wm_algorithm='tournier13',
        wm_N_candidate_voxels=300, gm_perc=0.02, csf_perc=0.1, **kwargs):
    """
    Heuristic approach to estimating the white matter, grey matter and CSF
    tissue response kernels [1]_, to be used in e.g. Multi-Tissue CSD [2]_. The
    method makes used of so-called 'optimal' thresholds between grey-scale
    images and segmentations [3]_, with iteratively refined binary thresholds
    based on an ad-hoc 'signal decay metric', to finally find candidate voxels
    from which to estimate the three tissue response kernels.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    data : NDarray,
        Measured diffusion signal array.
    wm_algorithm : string,
        selection of white matter response estimation algorithm:
        - 'tournier07': classic FA-based estimation,
        - 'tournier13': iterative peak-ratio based estimation.
    wm_N_candidate_voxels : positive integer,
        number of voxels to be included in the white matter response function.
        Default: 300 as done in [4]_.
    gm_perc : positive float between [0, 1],
        fraction of candidate voxels to use in grey matter response function.
        Default: 0.02 as done in [1]_.
    csf_perc : positive float between [0, 1],
        fraction of candidate voxels to use in CSF response function.
        Default: 0.1 as done in [1]_.
    kwargs : optional keyword arguments for WM algorithm,
        see white matter algorithms themselves for possible arguments.

    Returns
    -------
    [S0_wm, S0_gm, S0_csf] : list of floats,
        white matter, grey matter and csf responses.
    [TR2_wm_model, TR1_gm_model, TR1_csf_model]: list of
            TR2AnisotropicTissueResponseModel and
            2 TR1IsotropicTissueResponseModels,
        Modelfree signal representations of white/grey matter and csf.
    three_tissue_selection: array of size (x, y, z, 3),
        RGB mask of selected voxels used for white/grey matter and csf.

    References
    ----------
    .. [1] Dhollander, T.; Raffelt, D. & Connelly, A. Unsupervised 3-tissue
        response function estimation from single-shell or multi-shell diffusion
        MR data without a co-registered T1 image. ISMRM Workshop on Breaking
        the Barriers of Diffusion MRI, 2016, 5
    .. [2] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Determination of the appropriate b value and number of gradient
        directions for high-angular-resolution diffusion-weighted imaging."
        NMR in Biomedicine 26.12 (2013): 1775-1786.
    .. [3] Ridgway, Gerard R., et al. "Issues with threshold masking in
        voxel-based morphometry of atrophied brains." Neuroimage 44.1 (2009):
        99-111.
    .. [4] Tournier, J-Donald, Fernando Calamante, and Alan Connelly.
        "Determination of the appropriate b value and number of gradient
        directions for high-angular-resolution diffusion-weighted imaging."
        NMR in Biomedicine 26.12 (2013): 1775-1786.
    """
    # Create Signal Decay Metric (SDM)
    mean_b0 = np.mean(data[..., acquisition_scheme.b0_mask], axis=-1)
    SDM = signal_decay_metric(acquisition_scheme, data)

    # Make Mask
    b0_mask, mask = median_otsu(
        input_volume=data,
        vol_idx=np.where(acquisition_scheme.b0_mask)[0],
        median_radius=4, numpass=4)  # based on dipy default
    gtab = gtab_dmipy2dipy(acquisition_scheme)
    tenmod = dti.TensorModel(gtab)
    tenfit = tenmod.fit(b0_mask)
    fa = tenfit.fa
    mask_WM = fa > 0.2

    # Separate grey and CSF based on optimal threshold
    # take FA < 0.2 but inside brain mask.
    opt = optimal_threshold(SDM[np.all([fa < 0.2, mean_b0 > 0], axis=0)])
    mask_CSF = np.all([mean_b0 > 0, mask, fa < 0.2, SDM > opt], axis=0)
    mask_GM = np.all([mean_b0 > 0, mask, fa < 0.2, SDM < opt], axis=0)

    # Refine Mask, high WM SDM outliers above Q 3 +(Q 3 -Q 1 ) are removed.
    median_WM = np.median(SDM[mask_WM])
    Q1 = (SDM[mask_WM].min() + median_WM) / 2.0
    Q3 = (SDM[mask_WM].max() + median_WM) / 2.0
    SDM_upper_threshold = Q3 + (Q3 - Q1)
    mask_WM_refine = np.all([mask_WM, SDM < SDM_upper_threshold], axis=0)
    WM_outlier = np.all([mask_WM, SDM > SDM_upper_threshold], axis=0)

    # For both the voxels below and above the GM SDM median, optimal thresholds
    # [4] are computed and both parts closer to the initial GM median are
    # retained.
    SDM_GM = SDM[mask_GM]
    median_GM = np.median(SDM_GM)
    optimal_threshold_upper = optimal_threshold(SDM_GM[SDM_GM > median_GM])
    optimal_threshold_lower = optimal_threshold(SDM_GM[SDM_GM < median_GM])
    mask_GM_refine = np.all(
        [mask_GM,
         SDM > optimal_threshold_lower,
         SDM < optimal_threshold_upper], axis=0)

    # The high SDM outliers that were removed from the WM are reconsidered for
    # the CSF if they have higher SDM than the current minimal CSF SDM.
    SDM_CSF_min = SDM[mask_CSF].min()
    WM_outlier_to_include = np.all([WM_outlier, SDM > SDM_CSF_min], axis=0)
    mask_CSF_updated = np.any([mask_CSF, WM_outlier_to_include], axis=0)

    # An optimal threshold [4] is computed for the resulting CSF and only the
    # higher SDM valued voxels are retained.
    optimal_threshold_CSF = optimal_threshold(SDM[mask_CSF_updated])
    mask_CSF_refine = np.all(
        [mask_CSF_updated, SDM > optimal_threshold_CSF], axis=0)

    data_wm = data[mask_WM_refine]

    # for WM we use WM response selection algorithm
    response_wm_algorithm = _white_matter_response_algorithms[wm_algorithm]
    S0_wm, TR2_wm_model, indices_wm_selected = response_wm_algorithm(
        acquisition_scheme, data_wm, N_candidate_voxels=wm_N_candidate_voxels,
        **kwargs)

    # for GM, the voxels closest 2% to GM SDM median are selected.
    median_GM = np.median(SDM[mask_GM_refine])
    N_threshold = int(np.sum(mask_GM_refine) * gm_perc)
    indices_gm_selected = np.argsort(
        np.abs(SDM[mask_GM_refine] - median_GM))[:N_threshold]
    S0_gm, TR1_gm_model = estimate_TR1_isotropic_tissue_response_model(
        acquisition_scheme, data[mask_GM_refine][indices_gm_selected])

    # for GM, the 10% highest SDM valued voxels are selected.
    N_threshold = int(np.sum(mask_CSF_refine) * csf_perc)
    indices_csf_selected = np.argsort(SDM[mask_CSF_refine])[::-1][:N_threshold]
    S0_csf, TR1_csf_model = estimate_TR1_isotropic_tissue_response_model(
        acquisition_scheme, data[mask_CSF_refine][indices_csf_selected])

    # generate selected WM/GM/CSF response function voxels masks.
    pos_WM_refine = np.c_[np.where(mask_WM_refine)]
    mask_WM_selected = np.zeros_like(mask_WM_refine)
    pos_WM_selected = pos_WM_refine[indices_wm_selected]
    for pos in pos_WM_selected:
        mask_WM_selected[pos[0], pos[1], pos[2]] = 1

    pos_GM_refine = np.c_[np.where(mask_GM_refine)]
    mask_GM_selected = np.zeros_like(mask_GM_refine)
    pos_GM_selected = pos_GM_refine[indices_gm_selected]
    for pos in pos_GM_selected:
        mask_GM_selected[pos[0], pos[1], pos[2]] = 1

    pos_CSF_refine = np.c_[np.where(mask_CSF_refine)]
    mask_CSF_selected = np.zeros_like(mask_CSF_refine)
    pos_CSF_selected = pos_CSF_refine[indices_csf_selected]
    for pos in pos_CSF_selected:
        mask_CSF_selected[pos[0], pos[1], pos[2]] = 1

    three_tissue_selection = np.array(
        [mask_WM_selected, mask_GM_selected, mask_CSF_selected], dtype=float)
    three_tissue_selection = np.transpose(three_tissue_selection, (1, 2, 3, 0))

    return ([S0_wm, S0_gm, S0_csf],
            [TR2_wm_model, TR1_gm_model, TR1_csf_model],
            three_tissue_selection)


def signal_decay_metric(acquisition_scheme, data):
    """
    Estimation of the Signal Decay Metric (SDM) for the three-tissue tissue
    response kernel estimation [1]_. The metric is a simple division of the S0
    signal intensity by the b>0 shell's signal intensity - of the mean of their
    intensities if there are multiple shells.

    Parameters
    ----------
    acquisition_scheme : DmipyAcquisitionScheme instance,
        An acquisition scheme that has been instantiated using dMipy.
    data : NDarray,
        Measured diffusion signal array.

    Returns
    -------
    SDM : array of size data,
        Estimated Signal Decay Metric (SDK)

    References
    ----------
    .. [1] Dhollander, T.; Raffelt, D. & Connelly, A. Unsupervised 3-tissue
        response function estimation from single-shell or multi-shell diffusion
        MR data without a co-registered T1 image. ISMRM Workshop on Breaking
        the Barriers of Diffusion MRI, 2016, 5
    """
    mean_b0 = np.mean(data[..., acquisition_scheme.b0_mask], axis=-1)
    data_shape = data.shape[:-1]
    mean_dwi_shells = np.zeros(
        np.r_[data_shape, len(acquisition_scheme.unique_dwi_indices)])
    for i, index in enumerate(acquisition_scheme.unique_dwi_indices):
        shell_mask = acquisition_scheme.shell_indices == index
        mean_dwi_shells[..., i] = np.mean(data[..., shell_mask], axis=-1)

    SDM = np.zeros(data_shape)
    mask = np.min(np.concatenate((np.expand_dims(mean_b0, axis=-1),
                                  mean_dwi_shells), axis=-1), axis=-1) > 0
    ratio = np.log(mean_b0[mask, None] / mean_dwi_shells[mask])
    SDM[mask] = np.mean(ratio, axis=-1)

    if np.max(SDM) > 10 or np.min(SDM) < 0:
        warnings.warn(("The signal decay metric reached unrealistically " +
                      "high or negative values and was clipped to [0, 10]"),
                      RuntimeWarning)
        SDM = np.clip(SDM, 0, 10)
    return SDM


def optimal_threshold(data):
    """Optimal image threshold based on pearson correlation [1]_. The idea is
    that an 'optimal' mask of some arbitrary image data should be found by
    thresholding at a value that maximizes the pearson correlation between the
    original image and the mask, i.e:

    T* = argmax_T (\rho(data, data>T))
       = argmin_T -(\rho(data, data>T))

    This function estimates T* based on the second equation on arbitrary input
    arrays.

    Parameters
    ----------
    scalar_data: 1D array,
        scalar array to estimate an 'optimal' threshold on.

    Returns
    -------
    optimal_threshold: float,
        optimal threshold value that maximizes correlation between the original
        and masked data.

    References
    ----------
    .. [1] Ridgway, Gerard R., et al. "Issues with threshold masking in
        voxel-based morphometry of atrophied brains." Neuroimage 44.1 (2009):
        99-111.
    """
    min_bound = data.min()
    max_bound = data.max()
    eps = 1e-10
    optimal_threshold = brute(
        func=_cost_function,
        Ns=100,
        args=(data,),
        ranges=([min_bound + eps, max_bound - eps],))[0]
    return optimal_threshold


def _cost_function(threshold, image):
    "The cost function used by the optimal_threshold function."
    rho = -pearsonr(image, image > threshold)[0]
    return rho



from . import cylinder_models, plane_models
from ..utils import utils
from ..core.modeling_framework import ModelProperties
from ..core.signal_model_properties import AnisotropicSignalModelProperties
from ..core.constants import CONSTANTS
import numpy as np

__all__ = [
    'CC3CappedCylinderCallaghanApproximation',
]


class CC2CappedCylinderStejskalTannerApproximation(
        ModelProperties, AnisotropicSignalModelProperties):
    r""" The Stejskal-Tanner model for intra-cylindrical diffusion inside
    a capped cylinder with finite radius and length. The perpendicular
    diffusion is modelled after Soderman's solution for the disk [1]_. The
    parallel diffusion between planes has been implemented according to
    Balinov [2]_.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    diameter : float,
        capped cylinder (axon) diameter in meters.
    length : float,
        capped cylinder length in meters.

    References
    ----------
    .. [1] Soderman, Olle, and Bengt Jonsson. "Restricted diffusion in
        cylindrical geometry." Journal of Magnetic Resonance, Series A 117.1
        (1995): 94-97.
    .. [2] Balinov, Balin, et al. "The NMR self-diffusion method applied to
        restricted diffusion. Simulation of echo attenuation from molecules in
        spheres and between planes." Journal of Magnetic Resonance, Series A
        104.1 (1993): 17-25.
    """
    _required_acquisition_parameters = ['gradient_directions', 'qvalues']
    _model_type = 'CompartmentModel'

    def __init__(
        self,
        mu=None,
        diameter=None,
        length=None,
    ):
        self.mu = mu
        self.diameter = diameter
        self.length = length

        self._cylinder_model = (
            cylinder_models.C2CylinderStejskalTannerApproximation(
                mu=self.mu,
                diameter=self.diameter))
        self._plane_model = plane_models.P2PlaneStejskalTannerApproximation(
            diameter=length)

        self._parameter_ranges = self._cylinder_model._parameter_ranges.copy()
        self._parameter_ranges.update(self._plane_model._parameter_ranges)

        self._parameter_scales = self._cylinder_model._parameter_scales.copy()
        self._parameter_scales.update(self._plane_model._parameter_scales)

        self._parameter_types = self._cylinder_model._parameter_types.copy()
        self._parameter_types.update(self._plane_model._parameter_types)

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        n = acquisition_scheme.gradient_directions
        q = acquisition_scheme.qvalues

        diameter = kwargs.get('diameter', self.diameter)
        length = kwargs.get('length', self.length)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
        magnitude_perpendicular = np.linalg.norm(
            np.dot(mu_perpendicular_plane, n.T),
            axis=0
        )
        q_parallel = q * np.dot(n, mu)
        E_parallel = np.ones_like(q)
        q_nonzero = q_parallel > 0
        E_parallel[q_nonzero] = self._plane_model.plane_attenuation(
            q_parallel[q_nonzero], length)

        E_perpendicular = np.ones_like(q)
        q_perp = q * magnitude_perpendicular
        q_nonzero = q_perp > 0
        E_perpendicular[q_nonzero] = (
            self._cylinder_model.perpendicular_attenuation(
                q_perp[q_nonzero], diameter)
        )
        return E_parallel * E_perpendicular


class CC3CappedCylinderCallaghanApproximation(
        ModelProperties, AnisotropicSignalModelProperties):
    r""" The Callaghan model [1]_ - a cylinder with finite radius - for
    intra-axonal diffusion. The perpendicular diffusion is modelled
    after Callaghan's solution for the disk. The parallel diffusion of the
    capped cylinder is modelled using the same Callaghan approximation but
    between two parallel planes with a certain distance or 'length' between
    them.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    diameter : float,
        cylinder (axon) diameter in meters.
    length : float,
        cylinder length in meters.
    diffusion_intra : float,
        The diffusion constant of the water particles inside the cylinder.
        The default value is the approximate diffusivity of water inside axons
        as 1.7e-9 m^2/s.
    number_of_roots_cylinder : integer,
        number of roots for the cylinder Callaghan approximation.
    number_of_functions_cylinder : integer,
        number of functions for the cylinder Callaghan approximation.
    number_of_roots_plane : integer,
        number of roots for the plane Callaghan approximation.

    References
    ----------
    .. [1] Callaghan, Paul T. "Pulsed-gradient spin-echo NMR for planar,
        cylindrical, and spherical pores under conditions of wall
        relaxation." Journal of magnetic resonance, Series A 113.1 (1995):
        53-59.
    """
    _required_acquisition_parameters = [
        'gradient_directions', 'qvalues', 'tau']
    _model_type = 'CompartmentModel'

    def __init__(
        self,
        mu=None,
        diameter=None,
        length=None,
        diffusion_intra=CONSTANTS['water_in_axons_diffusion_constant'],
        number_of_roots_cylinder=20,
        number_of_functions_cylinder=50,
        number_of_roots_plane=40
    ):
        self.mu = mu
        self.diameter = diameter
        self.length = length
        self.diffusion_intra = diffusion_intra
        self.number_of_roots_cylinder = number_of_roots_cylinder
        self.number_of_functions_cylinder = number_of_functions_cylinder
        self.number_of_roots_plane = number_of_roots_plane

        self._cylinder_model = (
            cylinder_models.C3CylinderCallaghanApproximation(
                mu=self.mu,
                diameter=self.diameter,
                diffusion_perpendicular=self.diffusion_intra,
                number_of_roots=self.number_of_roots_cylinder,
                number_of_functions=self.number_of_functions_cylinder)
        )
        self._plane_model = plane_models.P3PlaneCallaghanApproximation(
            diameter=length,
            diffusion_constant=self.diffusion_intra,
            n_roots=self.number_of_roots_plane)

        self._parameter_ranges = self._cylinder_model._parameter_ranges.copy()
        self._parameter_ranges.update(self._plane_model._parameter_ranges)

        self._parameter_scales = self._cylinder_model._parameter_scales.copy()
        self._parameter_scales.update(self._plane_model._parameter_scales)

        self._parameter_types = self._cylinder_model._parameter_types.copy()
        self._parameter_types.update(self._plane_model._parameter_types)

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Calculates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        n = acquisition_scheme.gradient_directions
        q = acquisition_scheme.qvalues
        tau = acquisition_scheme.tau

        diameter = kwargs.get('diameter', self.diameter)
        length = kwargs.get('length', self.length)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
        magnitude_perpendicular = np.linalg.norm(
            np.dot(mu_perpendicular_plane, n.T),
            axis=0
        )
        q_parallel = q * np.dot(n, mu)
        E_parallel = np.ones_like(q)
        q_nonzero = q_parallel > 0
        E_parallel[q_nonzero] = self._plane_model.plane_attenuation(
            q_parallel[q_nonzero], tau[q_nonzero], length)
        E_perpendicular = np.ones_like(q)
        q_perp = q * magnitude_perpendicular

        q_nonzero = q_perp > 0
        E_perpendicular[q_nonzero] = (
            self._cylinder_model.perpendicular_attenuation(
                q_perp[q_nonzero], tau[q_nonzero], diameter)
        )
        return E_parallel * E_perpendicular



import numpy as np
from scipy.optimize import fmin_cobyla


class MultiTissueConvexOptimizer:
    """
    Secondary optimizer for including S0 tissue response values into the volume
    fraction estimation.

    Following the suggestion by [1]_, when including S0 responses, the
    volume fractions are no longer unity constrained. This means that the
    optimization of linear volume fractions and non-linear parameters is
    independent, and thus this secondary optimization is just a simple convex
    optimization on the volume fractions only.

    Parameters
    ----------
    model: dmipy multi-compartment model instance,
        dmipy initialized mc model.
    S0_tissue_responses: list,
        constains the positive S0 tissue responses that are associated with the
        tissue that each compartment model in the mc-model represents.

    References
    ----------
    .. [1] Dell'Acqua, Flavio, and J-Donald Tournier. "Modelling white matter
           with spherical deconvolution: How and why?." NMR in Biomedicine 32.4
           (2019): e3945.
    """

    def __init__(self, acquisition_scheme, model, S0_tissue_responses):
        self.acquisition_scheme = acquisition_scheme
        self.model = model
        self.S0_tissue_responses = S0_tissue_responses

    def cobyla_cost_function(self, fractions, phi, data):
        "Objective function of linear parameter estimation using COBYLA."
        E_hat = np.dot(phi, fractions)
        diff = data - E_hat
        objective = np.dot(diff, diff)
        return objective * 1e5

    def __call__(self, data, x0):
        params = x0 * self.model.scales_for_optimization
        params_dict = self.model.parameter_vector_to_parameters(params)
        phi = self.model(self.acquisition_scheme,
                         quantity="stochastic cost function", **params_dict)
        phi *= self.S0_tissue_responses

        if self.model.N_models == 1:
            vf_x0 = [1.]
        else:
            vf_x0 = x0[-self.model.N_models:]

        vf = fmin_cobyla(self.cobyla_cost_function, x0=vf_x0,
                         cons=[cobyla_positivity_constraint],
                         args=(phi, data),
                         maxfun=2000)
        return vf


def cobyla_positivity_constraint(volume_fractions, *args):
    "COBYLA positivity constraint on volume fractions"
    return volume_fractions - 0.001



# -*- coding: utf-8 -*-
'''
Document Module
'''
from __future__ import division
import pkg_resources
from os.path import join

import numpy as np
from scipy import stats
from scipy import special
from dipy.reconst.shm import real_sym_sh_mrtrix

from dmipy.utils import utils
from scipy import interpolate
from dmipy.core.modeling_framework import ModelProperties
from dipy.utils.optpkg import optional_package
from dipy.data import get_sphere, HemiSphere
sphere = get_sphere('symmetric724')
hemisphere = HemiSphere(phi=sphere.phi, theta=sphere.theta)

numba, have_numba, _ = optional_package("numba")

GRADIENT_TABLES_PATH = pkg_resources.resource_filename(
    'dmipy', 'data/gradient_tables'
)
SIGNAL_MODELS_PATH = pkg_resources.resource_filename(
    'dmipy', 'signal_models'
)
DATA_PATH = pkg_resources.resource_filename(
    'dmipy', 'data'
)
SPHERE_CARTESIAN = np.loadtxt(
    join(GRADIENT_TABLES_PATH, 'sphere_with_cap.txt')
)
SPHERE_SPHERICAL = utils.cart2sphere(SPHERE_CARTESIAN)
log_bingham_normalization_splinefit = np.load(
    join(DATA_PATH, "bingham_normalization_splinefit.npz"),
    encoding='bytes', allow_pickle=True)['arr_0']

inverse_sh_matrix_kernel = {
    sh_order: np.linalg.pinv(real_sym_sh_mrtrix(
        sh_order, hemisphere.theta, hemisphere.phi
    )[0]) for sh_order in np.arange(0, 15, 2)
}
BETA_SCALING = 1e-6

__all__ = [
    'get_sh_order_from_odi',
    'SD1Watson',
    'SD2Bingham',
    'DD1Gamma',
    'odi2kappa',
    'kappa2odi'
]


def get_sh_order_from_odi(odi):
    "Returns minimum sh_order to estimate spherical harmonics for given odi."
    odis = np.array([0.80606061, 0.46666667, 0.25333333,
                     0.15636364, 0.09818182, 0.06909091, 0.])
    sh_orders = np.arange(2, 15, 2)
    return sh_orders[np.argmax(odis < odi)]


class SD1Watson(ModelProperties):
    r""" The Watson spherical distribution model [1]_ [2]_.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    kappa : float,
        concentration parameter of the Watson distribution.

    References
    ----------
    .. [1] Kaden et al.
           "Parametric spherical deconvolution: inferring anatomical
            connectivity using diffusion MR imaging". NeuroImage (2007)
    .. [2] Zhang et al.
           "NODDI: practical in vivo neurite orientation dispersion and density
            imaging of the human brain". NeuroImage (2012)
    """

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'odi': (0.02, 0.99),
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'odi': 1.,
    }
    _parameter_types = {
        'mu': 'orientation',
        'odi': 'normal'
    }
    _model_type = 'SphericalDistribution'

    def __init__(self, mu=None, odi=None):
        self.mu = mu
        self.odi = odi

    def __call__(self, n, **kwargs):
        r""" The Watson spherical distribution model [1, 2].

        Parameters
        ----------
        n : array of shape(3) or array of shape(N x 3),
            sampled orientations of the Watson distribution.

        Returns
        -------
        Wn: float or array of shape(N),
            Probability density at orientations n, given mu and kappa.
        """
        odi = kwargs.get('odi', self.odi)
        mu = kwargs.get('mu', self.mu)

        kappa = odi2kappa(odi)
        mu_cart = utils.unitsphere2cart_1d(mu)
        numerator = np.exp(kappa * np.dot(n, mu_cart) ** 2)
        denominator = 4 * np.pi * special.hyp1f1(0.5, 1.5, kappa)
        Wn = numerator / denominator
        return Wn

    def spherical_harmonics_representation(self, sh_order=None, **kwargs):
        r""" The Watson spherical distribution model in spherical harmonics.
        The minimum order is automatically derived from numerical experiments
        to ensure fast function executation and accurate results.

        Parameters
        ----------
        sh_order : int,
            maximum spherical harmonics order to be used in the approximation.

        Returns
        -------
        watson_sh : array,
            spherical harmonics of Watson probability density.
        """
        odi = kwargs.get('odi', self.odi)
        mu = kwargs.get('mu', self.mu)
        if sh_order is None:
            sh_order = get_sh_order_from_odi(odi)

        watson_sf = self(hemisphere.vertices, mu=mu, odi=odi)
        sh_mat_inv = inverse_sh_matrix_kernel[sh_order]
        watson_sh = np.dot(sh_mat_inv, watson_sf)
        return watson_sh


class SD2Bingham(ModelProperties):
    r""" The Bingham spherical distribution model [1]_ [2]_ [3]_ using angles.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    psi : float,
        angle in radians of the bingham distribution around mu [0, pi].
    kappa : float,
        first concentration parameter of the Bingham distribution.
        defined as kappa = kappa1 - kappa3.
    beta : float,
        second concentration parameter of the Bingham distribution.
        defined as beta = kappa2 - kappa3. Bingham becomes Watson when beta=0.

    References
    ----------
    .. [1] Kaden et al.
           "Parametric spherical deconvolution: inferring anatomical
            connectivity using diffusion MR imaging". NeuroImage (2007)
    .. [2] Sotiropoulos et al.
           "Ball and rackets: inferring fiber fanning from
            diffusion-weighted MRI". NeuroImage (2012)
    .. [3] Tariq et al.
           "Bingham--NODDI: Mapping anisotropic orientation dispersion of
            neurites using diffusion MRI". NeuroImage (2016)
    """

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'psi': (0, np.pi),
        'odi': (0.02, 0.99),
        'beta_fraction': (0, 1)  # beta<=kappa in fact
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'psi': 1.,
        'odi': 1.,
        'beta_fraction': 1.
    }
    _parameter_types = {
        'mu': 'orientation',
        'psi': 'circular',
        'odi': 'normal',
        'beta_fraction': 'normal'
    }
    _model_type = 'SphericalDistribution'

    def __init__(self, mu=None, psi=None, odi=None, beta_fraction=None):
        self.mu = mu
        self.psi = psi
        self.odi = odi
        self.beta_fraction = beta_fraction

    def __call__(self, n, **kwargs):
        r""" The Watson spherical distribution model.

        Parameters
        ----------
        n : array of shape(3) or array of shape(N x 3),
            sampled orientations of the Watson distribution.

        Returns
        -------
        Bn: float or array of shape(N),
            Probability density at orientations n, given mu and kappa.
        """
        odi = kwargs.get('odi', self.odi)
        beta_fraction = kwargs.get('beta_fraction', self.beta_fraction)
        mu = kwargs.get('mu', self.mu)
        psi = kwargs.get('psi', self.psi)

        kappa = odi2kappa(odi)
        beta = beta_fraction * kappa

        mu_cart = utils.unitsphere2cart_1d(mu)

        R = utils.rotation_matrix_100_to_theta_phi_psi(mu[0], mu[1], psi)
        mu_beta = R.dot(np.r_[0., 1., 0.])
        numerator = _probability_bingham(kappa, beta, mu_cart, mu_beta, n)
        denominator = 4 * np.pi * self._get_normalization(kappa, beta)
        Bn = numerator / denominator
        return Bn

    def spherical_harmonics_representation(self, sh_order=None, **kwargs):
        r""" The Bingham spherical distribution model in spherical harmonics.
        The minimum order is automatically derived from numerical experiments
        to ensure fast function executation and accurate results.

        Parameters
        ----------
        sh_order : int,
            maximum spherical harmonics order to be used in the approximation.

        Returns
        -------
        bingham_sh : array,
            spherical harmonics of Bingham probability density.
        """
        odi = kwargs.get('odi', self.odi)
        beta_fraction = kwargs.get('beta_fraction', self.beta_fraction)
        mu = kwargs.get('mu', self.mu)
        psi = kwargs.get('psi', self.psi)
        if sh_order is None:
            sh_order = get_sh_order_from_odi(odi)

        bingham_sf = self(hemisphere.vertices, mu=mu, psi=psi, odi=odi,
                          beta_fraction=beta_fraction)

        sh_mat_inv = inverse_sh_matrix_kernel[sh_order]
        bingham_sh = np.dot(sh_mat_inv, bingham_sf)
        return bingham_sh

    def _get_normalization(self, kappa, beta):
        """
        The hyperconfluent function with matrix input is not available in
        python, so to normalize we estimated the bingham sphere function
        for kappa, beta in [0, 32] and estimated a 50x50 grid of its spherical
        means.

        Since the spherical mean of the bingham is similar to an exponential,
        we took its natural logarithm and fitted it to a 2D spline function.

        Below we use the fitted spline parameters in
        log_bingham_normalization_splinefit to interpolate the normalization
        for the distribution.

        code to generate the interpolation:

        from dipy.data import get_sphere, HemiSphere
        from dmipy.signal_models.spherical_mean import (
            estimate_spherical_mean_shell)
        import numpy as np
        sphere = get_sphere()
        n = HemiSphere(sphere.x, sphere.y, sphere.z).subdivide().vertices
        R = np.eye(3)
        norm_size = 50
        numerator = np.zeros(n.shape[0])
        norm_grid = np.ones((norm_size, norm_size))
        kappa_beta_range = np.linspace(0, 32, norm_size)
        for i in np.arange(norm_size):
            for j in np.arange(i + 1):
                Bdiag = np.diag(np.r_[kappa_beta_range[i],
                                kappa_beta_range[j],
                                0])
                B = R.dot(Bdiag).dot(R.T)
                for k, n_ in enumerate(n):
                    numerator[k] = np.exp(n_.dot(B).dot(n_))
                norm_grid[i, j] = norm_grid[j, i] = (
                    estimate_spherical_mean_shell(
                        numerator, n, sh_order=12))
        log_norm_grid = np.log(norm_grid)
        kappa_grid, beta_grid = np.meshgrid(kappa_beta_range, kappa_beta_range)
        from scipy import interpolate
        tck = interpolate.bisplrep(kappa_grid.ravel(),
                                   beta_grid.ravel(),
                                   log_norm_grid.ravel(), s=0)
        np.savez("bingham_normalization_splinefit.npz", tck)

        Parameters
        ----------
        kappa : float,
            first concentration parameter of the Bingham distribution.
            defined as kappa = kappa1 - kappa3.
        beta : float,
            second concentration parameter of the Bingham distribution.
            defined as beta = kappa2 - kappa3. Bingham becomes Watson when
            beta=0.

        Returns
        -------
        bingham_normalization: float
            spherical mean / normalization of the bingham distribution
        """
        log_norm = interpolate.bisplev(kappa, beta,
                                       log_bingham_normalization_splinefit)
        bingham_normalization = np.exp(log_norm)
        return bingham_normalization


class SD3SphericalHarmonics(ModelProperties):
    r"""A real-valued spherical harmonics distribution.

    Parameters
    ----------
    sh_order: int,
        maximum spherical harmonics order.
    sh_coeff: np.ndarray that must be of shape corresponding to sh_order.
        spherical harmonics coefficients of the distribution.
    """

    def __init__(self, sh_order, sh_coeff=None):
        self.sh_order = sh_order
        self.N_coeff = int((sh_order + 2) * (sh_order + 1) // 2)
        if sh_coeff is not None:
            if len(sh_coeff) != self.N_coeff:
                msg = 'if given, sh_coeff length must correspond to N_coeffs '\
                      'associated with sh_order ({} vs {}).'
                raise ValueError(msg.format(len(sh_coeff, self.N_coeff)))
        self.sh_coeff = sh_coeff

        self._parameter_ranges = {'sh_coeff': [
            [None, None] for i in range(self.N_coeff)]}
        self._parameter_scales = {'sh_coeff':
                                  np.ones(self.N_coeff, dtype=float)}
        self._parameter_cardinality = {'sh_coeff': self.N_coeff}
        self._parameter_types = {'sh_coeff': 'sh_coefficients'}
        self._parameter_optimization_flags = {'sh_coeff': True}

    def __call__(self, n, **kwargs):
        r"""Returns the sphere function at cartesian orientations n given
        spherical harmonic coefficients.

        Parameters
        ----------
        n : array of shape(N x 3),
            sampled orientations of the Watson distribution.

        Returns
        -------
        SHn: array of shape(N),
            Probability density at orientations n, given sh coeffs.
        """
        # calculate SHT matrix
        _, theta, phi = utils.cart2sphere(n).T
        SHT = real_sym_sh_mrtrix(self.sh_order, theta, phi)[0]
        # transform coefficients to sphere values
        sh_coeff = kwargs.get('sh_coeff', self.sh_coeff)
        SHn = SHT.dot(sh_coeff)
        return SHn

    def spherical_harmonics_representation(self, **kwargs):
        r"""Returns the spherical harmonic coefficients themselves.
        """
        return kwargs.get('sh_coeff', self.sh_coeff)


class DD1Gamma(ModelProperties):
    r"""A Gamma distribution of cylinder diameter for given alpha and beta
    parameters. NOTE: This is a distribution for axon DIAMETER and not SURFACE.
    To simulate the diffusion signal of an ensemble of gamma-distributed
    cylinders the probability still needs to be corrected for cylinder surface
    by multiplying by np.pi * radius ** 2 and renormalizing [1]_. Reason being
    that diffusion signals are generated by the volume of spins inside axons
    (cylinders), which is proportional to cylinder surface and not to diameter.

    Parameters
    ----------
    alpha : float,
        shape of the gamma distribution.
    beta : float,
        scale of the gamma distrubution. Different from Bingham distribution!

    References
    ----------
    .. [1] Assaf, Yaniv, et al. "AxCaliber: a method for measuring axon
        diameter distribution from diffusion MRI." Magnetic resonance in
        medicine 59.6 (2008): 1347-1354.
    """
    _parameter_ranges = {
        'alpha': (0.1, 30.),
        'beta': (1e-3, 2)
    }
    _parameter_scales = {
        'alpha': 1.,
        'beta': BETA_SCALING,
    }
    _parameter_types = {
        'alpha': 'normal',
        'beta': 'normal'
    }
    _model_type = 'SpatialDistribution'

    def __init__(self, alpha=None, beta=None, Nsteps=30,
                 normalization='standard'):
        self.alpha = alpha
        self.beta = beta
        self.Nsteps = Nsteps

        if normalization == 'standard':
            self.norm_func = self.unity
        elif normalization == 'plane':
            self.norm_func = self.length_plane
        elif normalization == 'cylinder':
            self.norm_func = self.surface_cylinder
        elif normalization == 'sphere':
            self.norm_func = self.volume_sphere
        else:
            msg = "Unknown normalization {}".format(normalization)
            raise ValueError(msg)
        self.calculate_sampling_start_and_end_points(self.norm_func)

    def length_plane(self, radius):
        "The distance normalization function for planes."
        return 2 * radius

    def surface_cylinder(self, radius):
        "The surface normalization function for cylinders."
        return np.pi * radius ** 2

    def volume_sphere(self, radius):
        "The volume normalization function for spheres."
        return (4. / 3.) * np.pi * radius ** 3

    def unity(self, radius):
        "The standard normalization for the Gamma distribution (none)."
        return np.ones(len(radius))

    def calculate_sampling_start_and_end_points(self, norm_func, gridsize=50):
        """
        For a given normalization function calculates the best start and end
        points to sample for all possible values of alpha, beta. This is done
        to make sure the function does not sample where the probability of
        basically zero.

        The function is based on first doing a dense sampling and then finding
        out which points need to be included to have sampled at least 99% of
        the area under the probability density curve.

        It sets two interpolator functions that can be called for any
        combination of alpha,beta and to return the appropriate start and end
        sampling points.

        Parameters
        ----------
        norm_func : normalization function,
            normalization of the model, depends on if it's a sphere/cylinder.
        gridsize : integer,
            value that decides how big the grid will be on which we define the
            start and end sampling points.
        """
        start_grid = np.ones([gridsize, gridsize])
        end_grid = np.ones([gridsize, gridsize])

        alpha_range = (np.array(self._parameter_ranges['alpha']) *
                       self._parameter_scales['alpha'])
        beta_range = (np.array(self._parameter_ranges['beta']) *
                      self._parameter_scales['beta'])

        alpha_linspace = np.linspace(alpha_range[0], alpha_range[1], gridsize)
        beta_linspace = np.linspace(beta_range[0], beta_range[1], gridsize)

        for i, alpha in enumerate(alpha_linspace):
            for j, beta in enumerate(beta_linspace):
                gamma_distribution = stats.gamma(alpha, scale=beta)
                outer_limit = (
                    gamma_distribution.mean() + 9 * gamma_distribution.std())
                x_grid = np.linspace(1e-8, outer_limit, 500)
                pdf = gamma_distribution.pdf(x_grid)
                pdf *= norm_func(x_grid)
                cdf = np.cumsum(pdf)
                cdf /= cdf.max()
                inverse_cdf = np.cumsum(pdf[::-1])[::-1]
                inverse_cdf /= inverse_cdf.max()
                end_grid[i, j] = x_grid[np.argmax(cdf > 0.995)]
                start_grid[i, j] = x_grid[np.argmax(inverse_cdf < 0.995)]
        start_grid = np.clip(start_grid, 1e-8, np.inf)
        end_grid = np.clip(end_grid, 1e-7, np.inf)

        alpha_grid, beta_grid = np.meshgrid(alpha_linspace, beta_linspace)

        self.start_interpolator = interpolate.bisplrep(alpha_grid.ravel(),
                                                       beta_grid.ravel(),
                                                       start_grid.T.ravel(),
                                                       kx=2, ky=2)

        self.end_interpolator = interpolate.bisplrep(alpha_grid.ravel(),
                                                     beta_grid.ravel(),
                                                     end_grid.T.ravel(),
                                                     kx=2, ky=2)

    def __call__(self, **kwargs):
        r"""
        Parameters
        ----------
        diameter : float or array, shape (N)
            cylinder (axon) diameter in meters.

        Returns
        -------
        Pgamma : float or array, shape (N)
            probability of cylinder diameter for given alpha and beta.
        """
        alpha = kwargs.get('alpha', self.alpha)
        beta = kwargs.get('beta', self.beta)

        gamma_dist = stats.gamma(alpha, scale=beta)
        start_point = interpolate.bisplev(alpha, beta, self.start_interpolator)
        end_point = interpolate.bisplev(alpha, beta, self.end_interpolator)
        start_point = max(start_point, 1e-8)
        radii = np.linspace(start_point, end_point, self.Nsteps)
        normalization = self.norm_func(radii)
        radii_pdf = gamma_dist.pdf(radii)
        radii_pdf_area = radii_pdf * normalization
        radii_pdf_normalized = (
            radii_pdf_area /
            np.trapz(x=radii, y=radii_pdf_area)
        )
        return radii, radii_pdf_normalized


def _probability_bingham(kappa, beta, mu, mu_beta, n):
    "Non-normalized probability of the Bingham distribution."
    return np.exp(kappa * np.dot(n, mu) ** 2 +
                  beta * np.dot(n, mu_beta) ** 2)


def odi2kappa(odi):
    "Calculates concentration (kappa) from orientation dispersion index (odi)."
    return 1. / np.tan(odi * (np.pi / 2.0))


def kappa2odi(kappa):
    "Calculates orientation dispersion index (odi) from concentration (kappa)."
    return (2. / np.pi) * np.arctan(1. / kappa)


if have_numba:
    get_sh_order_from_odi = numba.njit()(get_sh_order_from_odi)
    _probability_bingham = numba.njit()(_probability_bingham)



# -*- coding: utf-8 -*-
'''
Document Module
'''
from __future__ import division

import numpy as np
from scipy.special import erf

from ..utils import utils
from ..core.modeling_framework import ModelProperties
from ..core.signal_model_properties import (
    IsotropicSignalModelProperties, AnisotropicSignalModelProperties)
from dipy.utils.optpkg import optional_package

numba, have_numba, _ = optional_package("numba")

DIFFUSIVITY_SCALING = 1e-9
A_SCALING = 1e-12

__all__ = [
    'G1Ball',
    'G2Zeppelin',
    'G3TemporalZeppelin'
]


class G1Ball(ModelProperties, IsotropicSignalModelProperties):
    r""" The Ball model [1]_ - an isotropic Tensor with one diffusivity.

    Parameters
    ----------
    lambda_iso : float,
        isotropic diffusivity in m^2/s.

    References
    ----------
    .. [1] Behrens et al.
           "Characterization and propagation of uncertainty in
            diffusion-weighted MR imaging"
           Magnetic Resonance in Medicine (2003)
    """
    _required_acquisition_parameters = ['bvalues']

    _parameter_ranges = {
        'lambda_iso': (.1, 3)
    }
    _parameter_scales = {
        'lambda_iso': DIFFUSIVITY_SCALING
    }
    _parameter_types = {
        'lambda_iso': 'normal',
    }
    _model_type = 'CompartmentModel'

    def __init__(self, lambda_iso=None):
        self.lambda_iso = lambda_iso

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        lambda_iso = kwargs.get('lambda_iso', self.lambda_iso)
        E_ball = np.exp(-bvals * lambda_iso)
        return E_ball


class G2Zeppelin(ModelProperties, AnisotropicSignalModelProperties):
    r""" The Zeppelin model [1]_ - an axially symmetric Tensor - typically used
    for extra-axonal diffusion.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in m^2/s.
    lambda_perp : float,
        perpendicular diffusivity in m^2/s.

    Returns
    -------
    E_zeppelin : float or array, shape(N),
        signal attenuation.

    References
    ----------
    .. [1] Panagiotaki et al.
           "Compartment models of the diffusion MR signal in brain white
            matter: a taxonomy and comparison". NeuroImage (2012)
    """
    _required_acquisition_parameters = ['bvalues', 'gradient_directions']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3),
        'lambda_perp': (.1, 3)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING,
        'lambda_perp': DIFFUSIVITY_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal',
        'lambda_perp': 'normal',
    }
    _model_type = 'CompartmentModel'

    def __init__(self, mu=None, lambda_par=None, lambda_perp=None):
        self.mu = mu
        self.lambda_par = lambda_par
        self.lambda_perp = lambda_perp

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions
        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        lambda_perp = kwargs.get('lambda_perp', self.lambda_perp)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)
        E_zeppelin = _attenuation_zeppelin(
            bvals, lambda_par, lambda_perp, n, mu)
        return E_zeppelin

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme for
        Zeppelin model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the Zeppelin model for every acquisition shell.
        """
        bvals = acquisition_scheme.shell_bvalues[
            ~acquisition_scheme.shell_b0_mask]

        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        lambda_perp = kwargs.get('lambda_perp', self.lambda_perp)

        E_mean = np.ones_like(acquisition_scheme.shell_bvalues)
        if lambda_par > lambda_perp:  # use [kaden et al. 2016]
            exp_bl = np.exp(-bvals * lambda_perp)
            sqrt_bl = np.sqrt(bvals * (lambda_par - lambda_perp))
            E_mean_ = exp_bl * np.sqrt(np.pi) * erf(sqrt_bl) / (2 * sqrt_bl)
            E_mean[~acquisition_scheme.shell_b0_mask] = E_mean_
        else:  # estimate spherical mean using rotational harmonics
            rh_array = self.rotational_harmonics_representation(
                acquisition_scheme, **kwargs)
            E_mean[acquisition_scheme.unique_dwi_indices] = (
                rh_array[:, 0] / (2 * np.sqrt(np.pi))
            )
        return E_mean


class G3TemporalZeppelin(ModelProperties, AnisotropicSignalModelProperties):
    r"""
    The temporal Zeppelin model [1]_ - an axially symmetric Tensor - typically
    used to describe extra-axonal diffusion. The G3TemporalZeppelin differs
    from G2Zeppelin in that it has a time-dependent perpendicular parameter
    "A", which describe extra-axonal diffusion hindrance due to axon packing,
    and that lambda_perp is instead called lambda_inf, as it describes the
    perpendicular diffusivity when diffusion time is infinite.

    Parameters
    ----------
    mu : array, shape(2),
        angles [theta, phi] representing main orientation on the sphere.
        theta is inclination of polar angle of main angle mu [0, pi].
        phi is polar angle of main angle mu [-pi, pi].
    lambda_par : float,
        parallel diffusivity in 10^9 m^2/s.
    lambda_inf : float,
        bulk diffusivity constant 10^9 m^2/s.
    A: float,
        characteristic coefficient in 10^12 m^2

    Returns
    -------
    E_zeppelin : float or array, shape(N),
        signal attenuation.

    References
    ----------
    .. [1] Burcaw, L.M., Fieremans, E., Novikov, D.S., 2015. Mesoscopic
        structure of neuronal tracts from time-dependent diffusion.
        NeuroImage 114, 18.
    """
    _required_acquisition_parameters = [
        'bvalues', 'gradient_directions', 'delta', 'Delta']

    _parameter_ranges = {
        'mu': ([0, np.pi], [-np.pi, np.pi]),
        'lambda_par': (.1, 3),
        'lambda_inf': (.1, 3),
        'A': (0, 10)
    }
    _parameter_scales = {
        'mu': np.r_[1., 1.],
        'lambda_par': DIFFUSIVITY_SCALING,
        'lambda_inf': DIFFUSIVITY_SCALING,
        'A': A_SCALING
    }
    _parameter_types = {
        'mu': 'orientation',
        'lambda_par': 'normal',
        'lambda_inf': 'normal',
        'A': 'normal'
    }
    _model_type = 'CompartmentModel'

    def __init__(self, mu=None, lambda_par=None, lambda_inf=None, A=None):
        self.mu = mu
        self.lambda_par = lambda_par
        self.lambda_inf = lambda_inf
        self.A = A

    def __call__(self, acquisition_scheme, **kwargs):
        r'''
        Estimates the signal attenuation.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        attenuation : float or array, shape(N),
            signal attenuation
        '''
        bvals = acquisition_scheme.bvalues
        n = acquisition_scheme.gradient_directions
        delta = acquisition_scheme.delta
        Delta = acquisition_scheme.Delta

        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        lambda_inf = kwargs.get('lambda_inf', self.lambda_inf)
        A = kwargs.get('A', self.A)
        mu = kwargs.get('mu', self.mu)
        mu = utils.unitsphere2cart_1d(mu)

        R1 = mu
        R2 = utils.perpendicular_vector(R1)
        R3 = np.cross(R1, R2)
        R = np.c_[R1, R2, R3]

        E_zeppelin = np.ones_like(bvals)
        for i, (bval_, n_, delta_, Delta_) in enumerate(
            zip(bvals, n, delta, Delta)
        ):
            # lambda_perp and A must be in the same unit
            restricted_term = (
                A * (np.log(Delta_ / delta_) + 3 / 2.) / (Delta_ - delta_ / 3.)
            )
            D_perp = lambda_inf + restricted_term
            D_h = np.diag(np.r_[lambda_par, D_perp, D_perp])
            D = np.dot(np.dot(R, D_h), R.T)
            E_zeppelin[i] = np.exp(-bval_ * np.dot(n_, np.dot(n_, D)))
        return E_zeppelin

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme for
        Restricted Zeppelin model.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the Restricted Zeppelin model for every
            acquisition shell.
        """
        bvals = acquisition_scheme.shell_bvalues[
            ~acquisition_scheme.shell_b0_mask]
        delta = acquisition_scheme.shell_delta[
            ~acquisition_scheme.shell_b0_mask]
        Delta = acquisition_scheme.shell_Delta[
            ~acquisition_scheme.shell_b0_mask]
        lambda_par = kwargs.get('lambda_par', self.lambda_par)
        lambda_inf = kwargs.get('lambda_inf', self.lambda_inf)
        A = kwargs.get('A', self.A)
        E_mean = np.ones_like(acquisition_scheme.shell_bvalues)

        restricted_term = (
            A * (np.log(Delta / delta) + 3 / 2.) / (Delta - delta / 3.)
        )
        lambda_perp = lambda_inf + restricted_term
        if lambda_par > lambda_perp.max():  # use modified [kaden et al. 2016]
            exp_bl = np.exp(-bvals * lambda_perp)
            sqrt_bl = np.sqrt(bvals * (lambda_par - lambda_perp))
            E_mean[~acquisition_scheme.shell_b0_mask] = (
                exp_bl * np.sqrt(np.pi) * erf(sqrt_bl) / (2 * sqrt_bl))
        else:  # estimate spherical mean using rotational harmonics
            rh_array = self.rotational_harmonics_representation(
                acquisition_scheme, **kwargs)
            E_mean[acquisition_scheme.unique_dwi_indices] = (
                rh_array[:, 0] / (2 * np.sqrt(np.pi))
            )
        return E_mean


def _attenuation_zeppelin(bvals, lambda_par, lambda_perp, n, mu):
    "Signal attenuation for Zeppelin model."
    mu_perpendicular_plane = np.eye(3) - np.outer(mu, mu)
    magnitude_parallel = np.dot(n, mu)
    proj = np.dot(mu_perpendicular_plane, n.T)
    magnitude_perpendicular = np.sqrt(
        proj[0] ** 2 + proj[1] ** 2 + proj[2] ** 2)
    E_zeppelin = np.exp(-bvals *
                        (lambda_par * magnitude_parallel ** 2 +
                         lambda_perp * magnitude_perpendicular ** 2)
                        )
    return E_zeppelin


if have_numba:
    _attenuation_zeppelin = numba.njit()(_attenuation_zeppelin)



import numpy as np
from .gradient_conversions import (
    g_from_b, q_from_b, b_from_q, g_from_q, b_from_g, q_from_g)
from ..utils import utils
from ..utils.spherical_convolution import real_sym_rh_basis
from dipy.reconst.shm import real_sym_sh_mrtrix
from scipy.cluster.hierarchy import fcluster, linkage
from dipy.core.gradients import gradient_table, GradientTable
from warnings import warn


__all__ = [
    'get_sh_order_from_bval',
    'DmipyAcquisitionScheme',
    'RotationalHarmonicsAcquisitionScheme',
    'SphericalMeanAcquisitionScheme',
    'acquisition_scheme_from_bvalues',
    'acquisition_scheme_from_qvalues',
    'acquisition_scheme_from_gradient_strengths',
    'acquisition_scheme_from_schemefile',
    'unify_length_reference_delta_Delta',
    'calculate_shell_bvalues_and_indices',
    'check_acquisition_scheme',
    'gtab_dipy2dmipy',
    'gtab_dmipy2dipy'
]


def get_sh_order_from_bval(bval):
    "Estimates minimum sh_order to represent data of given b-value."
    bvals = np.r_[2.02020202e+08, 7.07070707e+08, 1.21212121e+09,
                  2.52525253e+09, 3.13131313e+09, 5.35353535e+09,
                  np.inf]
    sh_orders = np.arange(2, 15, 2)
    return sh_orders[np.argmax(bvals > bval)]


class DmipyAcquisitionScheme:
    """
    Class that calculates and contains all information needed to simulate and
    fit data using microstructure models.
    """

    def __init__(self, bvalues, gradient_directions, qvalues,
                 gradient_strengths, delta, Delta, TE,
                 min_b_shell_distance, b0_threshold):
        self.min_b_shell_distance = float(min_b_shell_distance)
        self.b0_threshold = float(b0_threshold)
        self.bvalues = bvalues.astype(float)
        self.b0_mask = self.bvalues <= b0_threshold
        self.number_of_b0s = np.sum(self.b0_mask)
        self.number_of_measurements = len(self.bvalues)
        self.gradient_directions = gradient_directions.astype(float)
        self.qvalues = None
        if qvalues is not None:
            self.qvalues = qvalues.astype(float)
        self.gradient_strengths = None
        if gradient_strengths is not None:
            self.gradient_strengths = gradient_strengths.astype(float)
        self.delta = None
        if delta is not None:
            self.delta = delta.astype(float)
        self.Delta = None
        if Delta is not None:
            self.Delta = Delta.astype(float)
        self.TE = None
        self.N_TE = 1  # default if not given
        if TE is not None:
            self.TE = TE.astype(float)
        self.tau = None
        if self.delta is not None and self.Delta is not None:
            self.tau = Delta - delta / 3.
        # if there are more then 1 measurement
        if self.number_of_measurements > 1:
            # we check if there are multiple unique delta-Delta combinations
            if self.TE is not None:
                deltas = np.c_[self.delta, self.Delta, self.TE]
            elif self.delta is not None and self.Delta is not None:
                deltas = np.c_[self.delta, self.Delta]
            elif self.delta is None and self.Delta is not None:
                deltas = np.c_[self.Delta]
            elif self.delta is not None and self.Delta is None:
                deltas = np.c_[self.delta]
            else:
                deltas = []

            if deltas == []:
                deltas = np.c_[np.zeros(len(self.bvalues))]
            unique_deltas = np.unique(deltas, axis=0)
            self.shell_indices = np.zeros(len(bvalues), dtype=int)
            self.shell_bvalues = []
            max_index = 0
            # for every unique combination we separate shells based on bvalue
            # reason for separation is that different combinations of
            # delta and Delta can result in the same b-value, which could
            # result in wrong classification of DWIs to unique shells.
            for unique_deltas_ in unique_deltas:
                delta_mask = np.all(deltas == unique_deltas_, axis=1)
                masked_bvals = bvalues[delta_mask]
                if len(masked_bvals) > 1:
                    shell_indices_, shell_bvalues_ = (
                        calculate_shell_bvalues_and_indices(
                            masked_bvals, min_b_shell_distance)
                    )
                else:
                    shell_indices_, shell_bvalues_ = np.array(0), masked_bvals
                self.shell_indices[delta_mask] = shell_indices_ + max_index
                self.shell_bvalues.append(shell_bvalues_)
                max_index = max(self.shell_indices + 1)
            self.shell_bvalues = np.hstack(self.shell_bvalues)
            self.shell_b0_mask = self.shell_bvalues <= b0_threshold

            first_indices = [
                np.argmax(self.shell_indices == ind)
                for ind in np.arange(self.shell_indices.max() + 1)]
            self.shell_qvalues = None
            if self.qvalues is not None:
                self.shell_qvalues = self.qvalues[first_indices]
            self.shell_gradient_strengths = None
            if self.gradient_strengths is not None:
                self.shell_gradient_strengths = (
                    self.gradient_strengths[first_indices])
            self.shell_delta = None
            if self.delta is not None:
                self.shell_delta = self.delta[first_indices]
            self.shell_Delta = None
            if self.Delta is not None:
                self.shell_Delta = self.Delta[first_indices]
            self.shell_TE = None
            if self.TE is not None:
                self.shell_TE = self.TE[first_indices]
                if (len(np.unique(self.TE)) != len(np.unique(
                        self.TE[self.b0_mask]))):
                    msg = "Not every TE shell has b0 measurements.\n"
                    msg += "This is required to properly normalize the signal."
                    msg += " Make sure the TE values for b0-measurements have "
                    msg += "not defaulted to 0 for example."
                    raise ValueError(msg)
                self.N_TE = len(self.shell_TE)
        # if for some reason only one measurement is given (for testing)
        else:
            self.shell_bvalues = self.bvalues
            self.shell_indices = np.r_[int(0)]
            if self.shell_bvalues > b0_threshold:
                self.shell_b0_mask = np.r_[False]
            else:
                self.shell_b0_mask = np.r_[True]
            self.shell_qvalues = self.qvalues
            self.shell_gradient_strengths = self.gradient_strengths
            self.shell_delta = self.delta
            self.shell_Delta = self.Delta
            self.shell_TE = TE

        # calculates observation matrices to convert spherical harmonic
        # coefficients to the positions on the sphere for every shell
        self.unique_b0_indices = np.unique(self.shell_indices[self.b0_mask])
        self.unique_dwi_indices = np.unique(self.shell_indices[~self.b0_mask])
        self.unique_shell_indices = np.unique(self.shell_indices)
        self.N_b0_shells = len(self.unique_b0_indices)
        self.N_dwi_shells = len(self.unique_dwi_indices)
        self.N_shells = len(self.unique_shell_indices)
        self.shell_sh_matrices = {}
        self.shell_sh_orders = {}
        for shell_index in self.unique_b0_indices:
            self.shell_sh_orders[shell_index] = 0
        for shell_index in self.unique_dwi_indices:
            shell_mask = self.shell_indices == shell_index
            bvecs_shell = self.gradient_directions[shell_mask]
            _, theta_, phi_ = utils.cart2sphere(bvecs_shell).T
            self.shell_sh_orders[shell_index] = get_sh_order_from_bval(
                self.shell_bvalues[shell_index])
            self.shell_sh_matrices[shell_index] = real_sym_sh_mrtrix(
                self.shell_sh_orders[shell_index], theta_, phi_)[0]
        # warning in case there are no b0 measurements
        if sum(self.b0_mask) == 0:
            msg = "No b0 measurements were detected. Check if the b0_threshold"
            msg += " option is high enough, or if there is a mistake in the "
            msg += "acquisition design."
            warn(msg)

        self.spherical_mean_scheme = SphericalMeanAcquisitionScheme(
            self.shell_bvalues,
            self.shell_qvalues,
            self.shell_gradient_strengths,
            self.shell_Delta,
            self.shell_delta)
        if len(self.unique_dwi_indices) > 0:
            self.rotational_harmonics_scheme = (
                RotationalHarmonicsAcquisitionScheme(self)
            )

    @property
    def print_acquisition_info(self):
        """
        prints a small summary of the acquisition scheme. Is useful to check if
        the function correctly separated the shells and if the input parameters
        were given in the right scale.
        """
        print("Acquisition scheme summary\n")
        print("total number of measurements: {}".format(
            self.number_of_measurements))
        print("number of b0 measurements: {}".format(self.number_of_b0s))
        print("number of DWI shells: {}\n".format(
            np.sum(~self.shell_b0_mask)))
        upper_line = "shell_index |# of DWIs |bvalue [s/mm^2] "
        upper_line += "|gradient strength [mT/m] |delta [ms] |Delta[ms]"
        upper_line += " |TE[ms]"
        print(upper_line)
        for ind in np.arange(max(self.shell_indices) + 1):
            if (self.shell_TE is not None and
                self.shell_delta is not None and
                    self.shell_Delta is not None):
                print(
                    "{:<12}|{:<10}|{:<16}|{:<25}|{:<11}|{:<10}|{:<5}".format(
                        str(ind), sum(self.shell_indices == ind),
                        int(self.shell_bvalues[ind] / 1e6),
                        int(1e3 * self.shell_gradient_strengths[ind]),
                        self.shell_delta[ind] * 1e3,
                        self.shell_Delta[ind] * 1e3, self.shell_TE[ind] * 1e3))
            elif (self.shell_TE is None and
                  self.shell_delta is not None and
                    self.shell_Delta is not None):
                print(
                    "{:<12}|{:<10}|{:<16}|{:<25}|{:<11}|{:<10}|{:<5}".format(
                        str(ind), sum(self.shell_indices == ind),
                        int(self.shell_bvalues[ind] / 1e6),
                        int(1e3 * self.shell_gradient_strengths[ind]),
                        self.shell_delta[ind] * 1e3,
                        self.shell_Delta[ind] * 1e3, 'N/A'))
            elif (self.shell_TE is None and
                  self.shell_delta is None and
                    self.shell_Delta is not None):
                print(
                    "{:<12}|{:<10}|{:<16}|{:<25}|{:<11}|{:<10}|{:<5}".format(
                        str(ind), sum(self.shell_indices == ind),
                        int(self.shell_bvalues[ind] / 1e6),
                        'N/A', 'N/A', self.shell_Delta[ind] * 1e3, 'N/A'))
            elif (self.shell_TE is None and
                  self.shell_delta is not None and
                    self.shell_Delta is None):
                print(
                    "{:<12}|{:<10}|{:<16}|{:<25}|{:<11}|{:<10}|{:<5}".format(
                        str(ind), sum(self.shell_indices == ind),
                        int(self.shell_bvalues[ind] / 1e6),
                        'N/A', self.shell_delta[ind] * 1e3, 'N/A', 'N/A'))
            elif (self.shell_TE is None and
                  self.shell_delta is None and
                    self.shell_Delta is None):
                print(
                    "{:<12}|{:<10}|{:<16}|{:<25}|{:<11}|{:<10}|{:<5}".format(
                        str(ind), sum(self.shell_indices == ind),
                        int(self.shell_bvalues[ind] / 1e6),
                        'N/A', 'N/A', 'N/A', 'N/A'))

    def to_schemefile(self, filename):
        """
        Exports acquisition scheme information in schemefile format, which can
        be used by the Camino Monte-Carlo simulator.

        Parameters
        ----------
        filename : string,
            location at which to save the schemefile.
        """
        TE_ = self.TE
        if TE_ is None:
            TE_ = self.Delta + 2 * self.delta + 0.001
        schemefile_data = np.hstack(
            [self.gradient_directions,
             self.gradient_strengths[:, None],
             self.Delta[:, None],
             self.delta[:, None],
             TE_[:, None]])
        header = "#g_x  g_y  g_z  |G| DELTA delta TE\n"
        header += "VERSION: STEJSKALTANNER"
        np.savetxt(filename, schemefile_data,
                   header=header, comments='')

    def visualise_acquisition_G_Delta_rainbow(
            self,
            Delta_start=None, Delta_end=None, G_start=None, G_end=None,
            bval_isolines=np.r_[0, 250, 1000, 2500, 5000, 7500, 10000, 14000],
            alpha_shading=0.6
    ):
        """This function visualizes a q-tau acquisition scheme as a function of
        gradient strength and pulse separation (big_delta). It represents every
        measurements at its G and big_delta position regardless of b-vector,
        with a background of b-value isolines for reference. It assumes there
        is only one unique pulse length (small_delta) in the acquisition
        scheme.

        Parameters
        ----------
        Delta_start : float,
            optional minimum big_delta that is plotted in seconds
        Delta_end : float,
            optional maximum big_delta that is plotted in seconds
        G_start : float,
            optional minimum gradient strength that is plotted in T/m
        G_end : float,
            optional maximum gradient strength taht is plotted in T/m
        bval_isolines : array,
            optional array of bvalue isolines that are plotted in background
            given in s/mm^2
        alpha_shading : float between [0-1]
            optional shading of the bvalue colors in the background
        """
        Delta = self.Delta  # in seconds
        delta = self.delta  # in seconds
        G = self.gradient_strengths  # in SI units T/m

        if len(np.unique(delta)) > 1:
            msg = "This acquisition has multiple small_delta values. "
            msg += "This visualization assumes there is only one small_delta."
            raise ValueError(msg)

        if Delta_start is None:
            Delta_start = 0.005
        if Delta_end is None:
            Delta_end = Delta.max() + 0.004
        if G_start is None:
            G_start = 0.
        if G_end is None:
            G_end = G.max() + .05

        Delta_ = np.linspace(Delta_start, Delta_end, 50)
        G_ = np.linspace(G_start, G_end, 50)
        Delta_grid, G_grid = np.meshgrid(Delta_, G_)
        bvals_ = b_from_g(G_grid.ravel(), delta[0], Delta_grid.ravel()) / 1e6
        bvals_ = bvals_.reshape(G_grid.shape)

        # local import because matplotlib is not in the strict requirements.
        import matplotlib.pyplot as plt
        plt.contourf(Delta_, G_, bvals_,
                     levels=bval_isolines,
                     cmap='rainbow', alpha=alpha_shading)
        cb = plt.colorbar(spacing="proportional")
        cb.ax.tick_params(labelsize=16)
        plt.scatter(Delta, G, c='k', s=25)

        plt.xlim(Delta_start, Delta_end)
        plt.ylim(G_start, G_end)
        cb.set_label('b-value ($s$/$mm^2$)', fontsize=18)
        plt.xlabel(r'Pulse Separation $\Delta$ [sec]', fontsize=18)
        plt.ylabel('Gradient Strength [T/m]', fontsize=18)

    def return_pruned_acquisition_scheme(self, shell_indices, data=None):
        """Returns pruned acquisition scheme and optionally also prunes data.

        Parameters
        ----------
        shell_indices: list of integers,
            the shell indices that correspond with the shells that should be
            returned. For the zeroth and second shell this is e.g. [0, 2]
        data: NDarray,
            DW-data that corresponds with the acquisition scheme. If it is
            given, then the data is pruned the same way as the acquisition
            scheme, meaning the pruned scheme and data can be used and fitted
            together again.

        Returns
        -------
        pruned_scheme: DmipyAcquisitionScheme object,
            the pruned acquisition scheme
        pruned_data: NDarray,
            the pruned data corresponding to the acquisition scheme.
        """
        booleans = []
        for index in shell_indices:
            booleans.append(self.shell_indices == index)
        mask = np.any(booleans, axis=0)

        bvals = self.bvalues[mask]
        gradient_directions = self.gradient_directions[mask]
        delta = self.delta[mask]
        Delta = self.Delta[mask]
        if self.TE is not None:
            TE = self.TE[mask]
        else:
            TE = None

        pruned_scheme = acquisition_scheme_from_bvalues(
            bvals, gradient_directions, delta, Delta, TE,
            self.min_b_shell_distance, self.b0_threshold)
        if data is None:
            return pruned_scheme
        else:
            pruned_data = data[..., mask]
            return pruned_scheme, pruned_data


class RotationalHarmonicsAcquisitionScheme:
    """
    AcquisitionScheme instance that contains the information necessary to
    calculate the rotational harmonics for a model for every acquisition shell.
    It is instantiated using a regular DmipyAcquisitionScheme and
    N_angular_samples determines how many samples are taken between mu=[0., 0.]
    and mu=[np.pi/2, 0.].

    Parameters
    ----------
    dmipy_acquisition_scheme: DmipyAcquisitionScheme instance
        An acquisition scheme that has been instantiated using dMipy.
    N_angular_samples: int
        Integer representing the number of angular samples per shell.
    """

    def __init__(self, dmipy_acquisition_scheme, N_angular_samples=10):
        self.Nsamples = N_angular_samples
        scheme = dmipy_acquisition_scheme

        thetas = np.linspace(0, np.pi / 2, N_angular_samples)
        r = np.ones(N_angular_samples)
        phis = np.zeros(N_angular_samples)
        angles = np.c_[r, thetas, phis]
        angles_cart = utils.sphere2cart(angles)

        b_all_shells = []
        Gdirs_all_shells = []
        delta_all_shells = []
        Delta_all_shells = []
        shell_indices = []
        for shell_index in scheme.unique_dwi_indices:
            b = scheme.shell_bvalues[shell_index]
            b_all_shells.append(np.tile(b, N_angular_samples))
            if scheme.shell_delta is not None:
                delta = scheme.shell_delta[shell_index]
                delta_all_shells.append(np.tile(delta, N_angular_samples))
            if scheme.shell_Delta is not None:
                Delta = scheme.shell_Delta[shell_index]
                Delta_all_shells.append(np.tile(Delta, N_angular_samples))
            Gdirs_all_shells.append(angles_cart)
            shell_indices.append(np.tile(shell_index, N_angular_samples))

        self.shell_indices = np.hstack(shell_indices)
        self.bvalues = np.hstack(b_all_shells)
        self.gradient_directions = np.vstack(Gdirs_all_shells)
        self.delta = None
        if scheme.shell_delta is not None:
            self.delta = np.hstack(delta_all_shells)
        self.Delta = None
        if scheme.shell_Delta is not None:
            self.Delta = np.hstack(Delta_all_shells)
        if self.delta is not None and self.Delta is not None:
            self.gradient_strengths = g_from_b(
                self.bvalues,
                self.delta,
                self.Delta)
            self.qvalues = q_from_g(
                self.gradient_strengths,
                self.delta)
            self.tau = self.Delta - self.delta / 3.0
        else:
            self.gradient_strengths = self.qvalues = self.tau = None
        self.b0_mask = np.tile(False, len(self.bvalues))
        self.shell_delta = scheme.shell_delta
        self.shell_Delta = scheme.shell_Delta
        self.unique_b0_indices = scheme.unique_b0_indices
        self.unique_shell_indices = scheme.unique_shell_indices
        self.unique_dwi_indices = scheme.unique_dwi_indices
        self.N_b0_shells = len(self.unique_b0_indices)
        self.N_dwi_shells = len(self.unique_dwi_indices)
        self.N_shells = len(self.unique_shell_indices)
        self.number_of_measurements = len(self.bvalues)

        self.shell_sh_matrices = {}
        self.shell_sh_orders = {}
        for shell_index in scheme.unique_dwi_indices:
            self.shell_sh_orders[shell_index] = int(
                scheme.shell_sh_orders[shell_index])
            self.shell_sh_matrices[shell_index] = real_sym_sh_mrtrix(
                self.shell_sh_orders[shell_index], thetas, phis)[0]

        self.inverse_rh_matrix = {
            rh_order: np.linalg.pinv(real_sym_rh_basis(
                rh_order, thetas, phis
            )) for rh_order in np.arange(0, 15, 2)
        }


class SphericalMeanAcquisitionScheme:
    "Acquisition scheme for isotropic spherical mean models."

    def __init__(self, bvalues, qvalues,
                 gradient_strengths, Deltas, deltas):
        self.bvalues = bvalues
        self.qvalues = qvalues
        self.gradient_strengths = gradient_strengths
        self.Delta = Deltas
        self.delta = deltas
        self.number_of_measurements = len(bvalues)


def acquisition_scheme_from_bvalues(
        bvalues, gradient_directions, delta=None, Delta=None, TE=None,
        min_b_shell_distance=50e6, b0_threshold=10e6):
    r"""
    Creates an acquisition scheme object from bvalues, gradient directions,
    pulse duration $\delta$ and pulse separation time $\Delta$.

    Parameters
    ----------
    bvalues: 1D numpy array of shape (Ndata)
        bvalues of the acquisition in s/m^2.
        e.g., a bvalue of 1000 s/mm^2 must be entered as 1000 * 1e6 s/m^2
    gradient_directions: 2D numpy array of shape (Ndata, 3)
        gradient directions array of cartesian unit vectors.
    delta: float or 1D numpy array of shape (Ndata)
        if float, pulse duration of every measurements in seconds.
        if array, potentially varying pulse duration per measurement.
    Delta: float or 1D numpy array of shape (Ndata)
        if float, pulse separation time of every measurements in seconds.
        if array, potentially varying pulse separation time per measurement.
    min_b_shell_distance : float
        minimum bvalue distance between different shells. This parameter is
        used to separate measurements into different shells, which is necessary
        for any model using spherical convolution or spherical mean.
    b0_threshold : float
        bvalue threshold for a measurement to be considered a b0 measurement.

    Returns
    -------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.
    """
    delta_, Delta_, TE_ = unify_length_reference_delta_Delta(
        bvalues, delta, Delta, TE)
    check_acquisition_scheme(
        bvalues, gradient_directions, delta_, Delta_, TE_)
    if delta is not None and Delta is not None:
        qvalues = q_from_b(bvalues, delta_, Delta_)
        gradient_strengths = g_from_b(bvalues, delta_, Delta_)
    else:
        qvalues = gradient_strengths = None
    return DmipyAcquisitionScheme(bvalues, gradient_directions, qvalues,
                                  gradient_strengths, delta_, Delta_, TE_,
                                  min_b_shell_distance, b0_threshold)


def acquisition_scheme_from_qvalues(
        qvalues, gradient_directions, delta, Delta, TE=None,
        min_b_shell_distance=50e6, b0_threshold=10e6):
    r"""
    Creates an acquisition scheme object from qvalues, gradient directions,
    pulse duration $\delta$ and pulse separation time $\Delta$.

    Parameters
    ----------
    qvalues: 1D numpy array of shape (Ndata)
        diffusion sensitization of the acquisition in 1/m.
        e.g. a qvalue of 10 1/mm must be entered as 10 * 1e3 1/m
    gradient_directions: 2D numpy array of shape (Ndata, 3)
        gradient directions array of cartesian unit vectors.
    delta: float or 1D numpy array of shape (Ndata)
        if float, pulse duration of every measurements in seconds.
        if array, potentially varying pulse duration per measurement.
    Delta: float or 1D numpy array of shape (Ndata)
        if float, pulse separation time of every measurements in seconds.
        if array, potentially varying pulse separation time per measurement.
    min_b_shell_distance : float
        minimum bvalue distance between different shells. This parameter is
        used to separate measurements into different shells, which is necessary
        for any model using spherical convolution or spherical mean.
    b0_threshold : float
        bvalue threshold for a measurement to be considered a b0 measurement.

    Returns
    -------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.
    """
    delta_, Delta_, TE_ = unify_length_reference_delta_Delta(
        qvalues, delta, Delta, TE)
    check_acquisition_scheme(
        qvalues, gradient_directions, delta_, Delta_, TE_)
    bvalues = b_from_q(qvalues, delta, Delta)
    gradient_strengths = g_from_q(qvalues, delta)
    return DmipyAcquisitionScheme(bvalues, gradient_directions, qvalues,
                                  gradient_strengths, delta_, Delta_, TE_,
                                  min_b_shell_distance, b0_threshold)


def acquisition_scheme_from_gradient_strengths(
        gradient_strengths, gradient_directions, delta, Delta, TE=None,
        min_b_shell_distance=50e6, b0_threshold=10e6):
    r"""
    Creates an acquisition scheme object from gradient strengths, gradient
    directions pulse duration $\delta$ and pulse separation time $\Delta$.

    Parameters
    ----------
    gradient_strengths: 1D numpy array of shape (Ndata)
        gradient strength of the acquisition in T/m.
        e.g., a gradient strength of 300 mT/m must be entered as 300 / 1e3 T/m
    gradient_directions: 2D numpy array of shape (Ndata, 3)
        gradient directions array of cartesian unit vectors.
    delta: float or 1D numpy array of shape (Ndata)
        if float, pulse duration of every measurements in seconds.
        if array, potentially varying pulse duration per measurement.
    Delta: float or 1D numpy array of shape (Ndata)
        if float, pulse separation time of every measurements in seconds.
        if array, potentially varying pulse separation time per measurement.
    min_b_shell_distance : float
        minimum bvalue distance between different shells. This parameter is
        used to separate measurements into different shells, which is necessary
        for any model using spherical convolution or spherical mean.
    b0_threshold : float
        bvalue threshold for a measurement to be considered a b0 measurement.

    Returns
    -------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.
    """
    delta_, Delta_, TE_ = unify_length_reference_delta_Delta(
        gradient_strengths, delta, Delta, TE)
    check_acquisition_scheme(gradient_strengths, gradient_directions,
                             delta_, Delta_, TE_)
    bvalues = b_from_g(gradient_strengths, delta, Delta)
    qvalues = q_from_g(gradient_strengths, delta)
    return DmipyAcquisitionScheme(bvalues, gradient_directions, qvalues,
                                  gradient_strengths, delta_, Delta_, TE_,
                                  min_b_shell_distance, b0_threshold)


def acquisition_scheme_from_schemefile(
        file_path, min_b_shell_distance=50e6, b0_threshold=10e6):
    r"""
    Created an acquisition scheme object from a Camino scheme file, containing
    gradient directions, strengths, pulse duration $\delta$ and pulse
    separation time $\Delta$ and TE.

    Parameters
    ----------
    file_path: string
        absolute file path to schemefile location
    min_b_shell_distance : float
        minimum bvalue distance between different shells. This parameter is
        used to separate measurements into different shells, which is necessary
        for any model using spherical convolution or spherical mean.
    b0_threshold : float
        bvalue threshold for a measurement to be considered a b0 measurement.

    Returns
    -------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.
    """
    skiprows = 0
    while True:
        try:
            scheme = np.loadtxt(file_path, skiprows=skiprows)
            break
        except ValueError:
            skiprows += 1

    bvecs = scheme[:, :3]
    bvecs[np.linalg.norm(bvecs, axis=1) == 0.] = np.r_[1., 0., 0.]
    G = scheme[:, 3]
    Delta = scheme[:, 4]
    delta = scheme[:, 5]
    TE = scheme[:, 6]
    return acquisition_scheme_from_gradient_strengths(
        G, bvecs, delta, Delta, TE, min_b_shell_distance, b0_threshold)


def unify_length_reference_delta_Delta(reference_array, delta, Delta, TE):
    """
    If either delta or Delta are given as float, makes them an array the same
    size as the reference array.

    Parameters
    ----------
    reference_array : array of size (Nsamples)
        typically b-values, q-values or gradient strengths.
    delta : float or array of size (Nsamples)
        pulse duration in seconds.
    Delta : float or array of size (Nsamples)
        pulse separation in seconds.
    TE : None, float or array of size (Nsamples)
        Echo time of the acquisition in seconds.

    Returns
    -------
    delta_ : array of size (Nsamples)
        pulse duration copied to be same size as reference_array
    Delta_ : array of size (Nsamples)
        pulse separation copied to be same size as reference_array
    TE_ : None or array of size (Nsamples)
        Echo time copied to be same size as reference_array
    """
    if delta is None:
        delta_ = delta
    elif isinstance(delta, float) or isinstance(delta, int):
        delta_ = np.tile(delta, len(reference_array))
    else:
        delta_ = delta.copy()
    if Delta is None:
        Delta_ = Delta
    elif isinstance(Delta, float) or isinstance(Delta, int):
        Delta_ = np.tile(Delta, len(reference_array))
    else:
        Delta_ = Delta.copy()
    if TE is None:
        TE_ = TE
    elif isinstance(TE, float) or isinstance(TE, int):
        TE_ = np.tile(TE, len(reference_array))
    else:
        TE_ = TE.copy()
    return delta_, Delta_, TE_


def calculate_shell_bvalues_and_indices(bvalues, max_distance=20e6):
    """
    Calculates which measurements belong to different acquisition shells.
    It uses scipy's linkage clustering algorithm, which uses the max_distance
    input as a limit of including measurements in the same cluster.

    For example, if bvalues were [1, 2, 3, 4, 5] and max_distance was 1, then
    all bvalues would belong to the same cluster.
    However, if bvalues were [1, 2, 4, 5] max max_distance was 1, then this
    would result in 2 clusters.

    Parameters
    ----------
    bvalues: 1D numpy array of shape (Ndata)
        bvalues of the acquisition in s/m^2.
    max_distance: float
        maximum b-value distance for a measurement to be included in the same
        shell.

    Returns
    -------
    shell_indices: 1D numpy array of shape (Ndata)
        array of integers, starting from 0, representing to which shell a
        measurement belongs. The number itself has no meaning other than just
        being different for different shells.
    shell_bvalues: 1D numpy array of shape (Nshells)
        array of the mean bvalues for every acquisition shell.
    """
    linkage_matrix = linkage(np.c_[bvalues])
    clusters = fcluster(linkage_matrix, max_distance, criterion='distance')
    shell_indices = np.empty_like(bvalues, dtype=int)
    cluster_bvalues = np.zeros((np.max(clusters), 2))
    for ind in np.unique(clusters):
        cluster_bvalues[ind - 1] = np.mean(bvalues[clusters == ind]), ind
    shell_bvalues, ordered_cluster_indices = (
        cluster_bvalues[cluster_bvalues[:, 0].argsort()].T)
    for i, ind in enumerate(ordered_cluster_indices):
        shell_indices[clusters == ind] = i
    return shell_indices, shell_bvalues


def check_acquisition_scheme(
        bqg_values, gradient_directions, delta, Delta, TE):
    "function to check the validity of the input parameters."
    if bqg_values.ndim > 1:
        msg = "b/q/G input must be a one-dimensional array. "
        msg += "Currently its dimensions is {}.".format(
            bqg_values.ndim
        )
        raise ValueError(msg)
    if len(bqg_values) != len(gradient_directions):
        msg = "b/q/G input and gradient_directions must have the same length. "
        msg += "Currently their lengths are {} and {}.".format(
            len(bqg_values), len(gradient_directions)
        )
        raise ValueError(msg)
    if delta is not None:
        if len(bqg_values) != len(delta):
            msg = "b/q/G input and delta must have the same length. "
            msg += "Currently their lengths are {} and {}.".format(
                len(bqg_values), len(delta)
            )
            raise ValueError(msg)
        if delta.ndim > 1:
            msg = "delta must be one-dimensional array. "
            msg += "Currently its dimension is {}".format(
                delta.ndim
            )
            raise ValueError(msg)
        if np.min(delta) < 0:
            msg = "delta must be zero or positive. "
            msg += "Currently its minimum value is {}.".format(
                np.min(delta)
            )
            raise ValueError(msg)
    if Delta is not None:
        if len(bqg_values) != len(Delta):
            msg = "b/q/G input and Delta must have the same length. "
            msg += "Currently their lengths are {} and {}.".format(
                len(bqg_values), len(Delta)
            )
            raise ValueError(msg)
        if Delta.ndim > 1:
            msg = "Delta must be one-dimensional array. "
            msg += "Currently its dimension is {}.".format(
                Delta.ndim
            )
            raise ValueError(msg)
        if np.min(Delta) < 0:
            msg = "Delta must be zero or positive. "
            msg += "Currently its minimum value is {}.".format(
                np.min(Delta)
            )
            raise ValueError(msg)

    if gradient_directions.ndim != 2 or gradient_directions.shape[1] != 3:
        msg = "gradient_directions n must be two dimensional array of shape "
        msg += "[N, 3]. Currently its shape is {}.".format(
            gradient_directions.shape)
        raise ValueError(msg)
    if np.min(bqg_values) < 0.:
        msg = "b/q/G input must be zero or positive. "
        msg += "Minimum value found is {}.".format(bqg_values.min())
        raise ValueError(msg)
    gradient_norms = np.linalg.norm(gradient_directions, axis=1)
    zero_norms = gradient_norms == 0.
    if not np.all(abs(gradient_norms[~zero_norms] - 1.) < 0.001):
        msg = "gradient orientations n are not unit vectors. "
        raise ValueError(msg)
    if TE is not None and len(TE) != len(bqg_values):
        msg = "If given, TE must be same length b/q/G input."
        msg += "Currently their lengths are {} and {}.".format(
            len(TE), len(gradient_directions)
        )


def gtab_dipy2dmipy(dipy_gradient_table, min_b_shell_distance=50e6,
                    b0_threshold=10e6):
    """Converts a dipy gradient_table to a dmipy acquisition_scheme.
    If no big_delta or small_delta is defined in the gradient table, then None
    is passed to the DmipyAcquisitionScheme for these fields, and no models
    can be used that need this information.

    Parameters
    ----------
    dipy_gradient_table: dipy GradientTable instance,
        object that contains bvals, bvecs, pulse separation and duration
        information.
    min_b_shell_distance : float
        minimum bvalue distance between different shells. This parameter is
        used to separate measurements into different shells, which is necessary
        for any model using spherical convolution or spherical mean.
    b0_threshold : float
        bvalue threshold for a measurement to be considered a b0 measurement.

    Returns
    -------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.

    """
    if not isinstance(dipy_gradient_table, GradientTable):
        msg = "Input must be a dipy GradientTable object. "
        raise ValueError(msg)
    bvals = dipy_gradient_table.bvals * 1e6
    bvecs = dipy_gradient_table.bvecs
    delta = dipy_gradient_table.small_delta
    Delta = dipy_gradient_table.big_delta

    if delta is None or Delta is None:
        msg = "pulse_separation (big_delta) or pulse_duration (small_delta) "
        msg += "are not defined in the Dipy gtab. This means the resulting "
        msg += "DmipyAcquisitionScheme cannot be used with CompartmentModels "
        msg += "that need these."
        warn(msg)

    gtab_dmipy = acquisition_scheme_from_bvalues(
        bvalues=bvals, gradient_directions=bvecs, delta=delta, Delta=Delta,
        min_b_shell_distance=min_b_shell_distance, b0_threshold=b0_threshold)
    return gtab_dmipy


def gtab_dmipy2dipy(dmipy_gradient_table):
    """Converts a dmipy acquisition scheme to a dipy gradient_table.

    Parameters
    ----------
    DmipyAcquisitionScheme: acquisition scheme object
        contains all information of the acquisition scheme to be used in any
        microstructure model.

    Returns
    -------
    dipy_gradient_table: dipy GradientTable instance,
        object that contains bvals, bvecs, pulse separation and duration
        information.
    """
    if not isinstance(dmipy_gradient_table, DmipyAcquisitionScheme):
        msg = "Input must be a DmipyAcquisitionScheme object. "
        raise ValueError(msg)
    bvals = dmipy_gradient_table.bvalues / 1e6
    bvecs = dmipy_gradient_table.gradient_directions
    delta = dmipy_gradient_table.delta
    Delta = dmipy_gradient_table.Delta

    if delta is None:
        pass  # leave delta undefined in dipy gtab
    elif len(np.unique(delta)) > 1:
        msg = "Cannot create Dipy GradientTable for Acquisition schemes with "
        msg += "multiple delta (pulse duration) values, due to current "
        msg += "limitations of Dipy GradientTables."
        raise ValueError(msg)
    elif len(np.unique(delta)) == 1:
        delta = delta[0]

    if Delta is None:
        pass  # leave Delta undefined in dipy gtab
    elif len(np.unique(Delta)) > 1:
        msg = "Cannot create Dipy GradientTable for Acquisition schemes with "
        msg += "multiple Delta (pulse sepration) values, due to current "
        msg += "limitations of Dipy GradientTables."
        raise ValueError(msg)
    elif len(np.unique(Delta)) == 1:
        Delta = Delta[0]

    dipy_gradient_table = gradient_table(
        bvals=bvals, bvecs=bvecs, small_delta=delta, big_delta=Delta)
    return dipy_gradient_table



import numpy as np
from ..utils.construct_observation_matrix import construct_model_based_A_matrix


class AnisotropicSignalModelProperties:
    def __init__(self):
        pass

    def __call__(self):
        raise NotImplementedError()

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        kwargs.update({'mu': [0., 0.]})
        E_kernel_sf = self(rh_scheme, **kwargs)
        E_reshaped = E_kernel_sf.reshape([-1, rh_scheme.Nsamples])
        max_sh_order = max(rh_scheme.shell_sh_orders.values())
        rh_array = np.zeros((len(E_reshaped), max_sh_order // 2 + 1))

        for i, (shell_index, sh_order) in enumerate(
                rh_scheme.shell_sh_orders.items()):
            rh_array[i, :sh_order // 2 + 1] = (
                np.dot(
                    rh_scheme.inverse_rh_matrix[sh_order],
                    E_reshaped[i])
            )
        return rh_array

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the model for every acquisition shell.
        """
        E_mean = np.ones_like(acquisition_scheme.shell_bvalues)
        rh_array = self.rotational_harmonics_representation(
            acquisition_scheme, **kwargs)
        E_mean[acquisition_scheme.unique_dwi_indices] = (
            rh_array[:, 0] / (2 * np.sqrt(np.pi))
        )
        return E_mean

    def convolution_kernel_matrix(self, acquisition_scheme, lmax, **kwargs):
        """Constructs the multi-shell observation matrix from spherical_harmonics
        to DWIs. Follows the notation of Eq. (2) in [1]_.

        The dmipy acquisition_scheme object contains all the information on
        which DWIs belong to which acquisition shells, what are the maximum
        spherical harmonics order used for each shell, and the observation
        matrix that maps the DWIs of each shell to a spherical harmonics
        representation.

        The dmipy model must be have all parameters fixed to be able to
        generate the rotational harmonics of each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
                An acquisition scheme that has been instantiated using dmipy.
        lmax: even positive integer,
            even maximum spherical harmonics order of the to-be-estimated FOD.

        Returns
        -------
        Ams: array of size (N_DWIs, N_sh_coef),
            observation matrix to map spherical harmonics to DWIs.

        References
        ----------
        .. [1] Jeurissen, Ben, et al. "Multi-tissue constrained spherical
            deconvolution for improved analysis of multi-shell diffusion MRI
            data." NeuroImage 103 (2014): 411-426.
        """
        model_rh = self.rotational_harmonics_representation(
            acquisition_scheme, **kwargs)
        return construct_model_based_A_matrix(
            acquisition_scheme, model_rh, lmax)


class IsotropicSignalModelProperties:
    def __init__(self):
        pass

    def __call__(self):
        raise NotImplementedError()

    def rotational_harmonics_representation(
            self, acquisition_scheme, **kwargs):
        r""" The rotational harmonics of the model, such that Y_lm = Yl0.
        Axis aligned with z-axis to be used as kernel for spherical
        convolution. Returns an array with rotational harmonics for each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        rh_array : array, shape(Nshells, N_rh_coef),
            Rotational harmonics coefficients for each shell.
        """
        rh_scheme = acquisition_scheme.rotational_harmonics_scheme
        E_kernel_sf = self(rh_scheme, **kwargs)
        E_reshaped = E_kernel_sf.reshape([-1, rh_scheme.Nsamples])
        rh_array = np.zeros((len(E_reshaped), 1))

        for i, sh_order in enumerate(rh_scheme.shell_sh_orders):
            rh_array[i, :sh_order // 2 + 1] = (
                np.dot(
                    rh_scheme.inverse_rh_matrix[0],
                    E_reshaped[i])
            )
        return rh_array

    def spherical_mean(self, acquisition_scheme, **kwargs):
        """
        Estimates spherical mean for every shell in acquisition scheme.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
            An acquisition scheme that has been instantiated using dMipy.
        kwargs: keyword arguments to the model parameter values,
            Is internally given as **parameter_dictionary.

        Returns
        -------
        E_mean : float,
            spherical mean of the model for every acquisition shell.
        """
        return self(acquisition_scheme.spherical_mean_scheme, **kwargs)

    def convolution_kernel_matrix(self, acquisition_scheme, lmax, **kwargs):
        """Constructs the multi-shell observation matrix from spherical_harmonics
        to DWIs. Follows the notation of Eq. (2) in [1]_.

        The dmipy acquisition_scheme object contains all the information on
        which DWIs belong to which acquisition shells, what are the maximum
        spherical harmonics order used for each shell, and the observation
        matrix that maps the DWIs of each shell to a spherical harmonics
        representation.

        The dmipy model must be have all parameters fixed to be able to
        generate the rotational harmonics of each shell.

        Parameters
        ----------
        acquisition_scheme : DmipyAcquisitionScheme instance,
                An acquisition scheme that has been instantiated using dmipy.
        lmax: even positive integer,
            even maximum spherical harmonics order of the to-be-estimated FOD.

        Returns
        -------
        Ams: array of size (N_DWIs, N_sh_coef),
            observation matrix to map spherical harmonics to DWIs.

        References
        ----------
        .. [1] Jeurissen, Ben, et al. "Multi-tissue constrained spherical
            deconvolution for improved analysis of multi-shell diffusion MRI
            data." NeuroImage 103 (2014): 411-426.
        """

        model_rh = self.rotational_harmonics_representation(
            acquisition_scheme, **kwargs)

        Ncoef = int((lmax + 2) * (lmax + 1) // 2)
        Ams = np.zeros([acquisition_scheme.number_of_measurements, Ncoef])

        Ams[:, :1] = construct_model_based_A_matrix(
            acquisition_scheme, model_rh, 0)
        return Ams



import numpy as np
from dipy.reconst.shm import real_sym_sh_mrtrix
from . import utils

__all__ = [
    'estimate_spherical_mean_multi_shell',
    'estimate_spherical_mean_shell'
]


def estimate_spherical_mean_multi_shell(E_attenuation, acquisition_scheme,
                                        sh_order=6):
    r""" Estimates the spherical mean per shell of multi-shell acquisition.
    Uses spherical harmonics to do the estimation.

    Parameters
    ----------
    E_attenuation : array, shape (N),
        signal attenuation.
    bvecs : array, shape (N x 3),
        x, y, z components of cartesian unit b-vectors.
    b_shell_indices : array, shape (N)
        array of integers indicating which measurement belongs to which shell.
        0 should be used for b0 measurements, 1 for the lowest b-value shell,
        2 for the second lowest etc.

    Returns
    -------
    E_mean : array, shape (number of b-shells)
        spherical means of the signal attenuation per shell of the. For
        example, if there are three shells in the acquisition then the array
        is of length 3.
    """
    shell_indices = acquisition_scheme.shell_indices
    gradient_directions = acquisition_scheme.gradient_directions
    unique_indices = acquisition_scheme.unique_shell_indices

    E_mean = np.ones(acquisition_scheme.N_shells)
    for b_shell_index in unique_indices:  # per shell
        shell_mask = shell_indices == b_shell_index
        E_shell = E_attenuation[shell_mask]
        if acquisition_scheme.shell_b0_mask[b_shell_index]:
            E_mean[b_shell_index] = np.mean(E_shell)
        else:
            sh_mat = acquisition_scheme.shell_sh_matrices[b_shell_index]
            bvecs_shell = gradient_directions[shell_mask]
            E_mean[b_shell_index] = estimate_spherical_mean_shell(
                E_shell, bvecs_shell, sh_order, sh_mat)
    return E_mean


def estimate_spherical_mean_shell(
        E_shell, bvecs_shell, sh_order=6, sh_mat=None):
    """ Estimates spherical mean of a shell of measurements using
    spherical harmonics.
    The spherical mean is contained only in the Y00 spherical harmonic, as long
    as the basis expansion order is sufficient to capture the spherical signal.

    Parameters
    ----------
    E_shell : array, shape(N),
        signal attenuation values.
    bvecs_shell : array, shape (N x 3),
        Cartesian unit vectors describing the orientation of the signal
        attenuation values.
    sh_order : integer,
        maximum spherical harmonics order. It needs to be high enough to
        describe the spherical profile of the signal attenuation. The order 6
        is sufficient to describe a stick at b-values up to 10,000 s/mm^2.
    sh_mat : array of size (N_bvecs, N_coefficients),
        possibly precomputed spherical harmonics transform matrix.

    Returns
    -------
    E_mean : float,
        spherical mean of the signal attenuation.
    """
    if sh_mat is None:
        _, theta_, phi_ = utils.cart2sphere(bvecs_shell).T
        sh_mat = real_sym_sh_mrtrix(sh_order, theta_, phi_)[0]
    sh_mat_inv = np.linalg.pinv(sh_mat)
    E_sh_coef = np.dot(sh_mat_inv, E_shell)
    # Integral of sphere is 1 / (4 * np.pi)
    # Integral of Y00 spherical harmonic is 2 * np.sqrt(np.pi)
    # Multiplication results in normalization of 1 / (2 * np.sqrt(np.pi))
    E_mean = E_sh_coef[0] / (2 * np.sqrt(np.pi))
    return E_mean



from dmipy.signal_models.gaussian_models import G1Ball
from dmipy.core.modeling_framework import MultiCompartmentModel
import numpy as np
from time import time


def ivim_Dstar_fixed(acquisition_scheme, data, mask=None, Dstar_value=7e-9,
                     solver='brute2fine', **fit_args):
    """
    Implementation of second best performing IVIM algorithm following [1]_.
    Basically, it is just a non-linear least squares fit with fixing the
    blood diffusivity Dstar to 7e-3 mm^2/s. This value apparently improves the
    stability of the fit (in healthy volunteers) [2]_.

    The optimization range for the tissue diffusivity is set to
    [0.5 - 6]e-3 mm^2/s to improve precision [3]_.

    In the fitted ivim_fit model, partial_volume_0 and G1Ball_1_lambda_iso
    represent the tissue fraction and diffusivity, and partial_volume_1 and
    G1Ball_2_lambda_iso represent the blood fraction and diffusivity.

    Parameters
    ----------
    acquisition_scheme: Dmipy AcquisitionScheme instance,
        acquisition scheme containing all the information of the ivim
        acquisition.
    data: ND-array of shape (Nx, ..., N_DWI),
        measured data corresponding to the acquisition scheme.
    mask : (N-1)-dimensional integer/boolean array of size (N_x, N_y, ...),
        Optional mask of voxels to be included in the optimization.
    Dstar_value: float,
        the fixed Dstar blood diffusivity value. Default: 7e-9 m^2/s [2]_.
    solver: float,
        which solver to use for the algorithm. Default: 'brute2fine'.
    fit_args: other keywords that are passed to the optimizer

    Returns
    -------
    ivim_fit: Dmipy FittedMultiCompartmentModel instance,
        contains the fitted IVIM parameters.

    References
    ----------
    .. [1] Gurney-Champion, O. J., Klaassen, R., Froeling, M., Barbieri, S.,
        Stoker, J., Engelbrecht, M. R., ... & Nederveen, A. J. (2018).
        Comparison of six fit algorithms for the intra-voxel incoherent motion
        model of diffusion-weighted magnetic resonance imaging data of
        pancreatic cancer patients. PloS one, 13(4), e0194590.
    .. [2] Gurney-Champion OJ, Froeling M, Klaassen R, Runge JH, Bel A, Van
        Laarhoven HWM, et al. Minimizing the Acquisition Time for Intravoxel
        Incoherent Motion Magnetic Resonance Imaging Acquisitions in the Liver
        and Pancreas. Invest Radiol. 2016;51: 211–220.
    .. [3] Park HJ, Sung YS, Lee SS, Lee Y, Cheong H, Kim YJ, et al. Intravoxel
        incoherent motion diffusion-weighted MRI of the abdomen: The effect of
        fitting algorithms on the accuracy and reliability of the parameters.
        J Magn Reson Imaging. 2017;45: 1637–1647.
    """
    start = time()

    if fit_args is None:
        fit_args = {}

    print('Starting IVIM Dstar-fixed algorithm.')
    ivim_mod = MultiCompartmentModel([G1Ball(), G1Ball()])
    ivim_mod.set_fixed_parameter(
        'G1Ball_2_lambda_iso', Dstar_value)  # following [2]
    ivim_mod.set_parameter_optimization_bounds(
        'G1Ball_1_lambda_iso', [.5e-9, 6e-9])  # following [3]
    ivim_fit = ivim_mod.fit(
        acquisition_scheme=acquisition_scheme,
        data=data,
        mask=mask,
        solver=solver,
        **fit_args)
    computation_time = time() - start
    N_voxels = np.sum(ivim_fit.mask)
    msg = 'IVIM Dstar-fixed optimization of {0:d} voxels'.format(N_voxels)
    msg += ' complete in {0:.3f} seconds'.format(computation_time)
    print(msg)
    return ivim_fit
